#%%
import xarray as xr
import numpy as np

# List of dataset file paths, to add more fields, just add path to .nc file
# Data must already be on a common grid (example data is 180x360x41x12)
file_paths = [
    '/home/mv23682/Documents/Abil/studies/wiseman2024/env_data_processing/regridded_data/temperature.nc',
    '/home/mv23682/Documents/Abil/studies/wiseman2024/env_data_processing/regridded_data/sio4.nc',
    '/home/mv23682/Documents/Abil/studies/wiseman2024/env_data_processing/regridded_data/po4.nc',
    '/home/mv23682/Documents/Abil/studies/wiseman2024/env_data_processing/regridded_data/no3.nc',
    '/home/mv23682/Documents/Abil/studies/wiseman2024/env_data_processing/regridded_data/o2.nc',
    '/home/mv23682/Documents/Abil/studies/wiseman2024/env_data_processing/regridded_data/DIC.nc',
    '/home/mv23682/Documents/Abil/studies/wiseman2024/env_data_processing/regridded_data/TA.nc',
    '/home/mv23682/Documents/Abil/studies/wiseman2024/env_data_processing/regridded_data/PAR.nc',
    '/home/mv23682/Documents/Abil/studies/wiseman2024/env_data_processing/regridded_data/MLD_SEANOE.nc',
    '/home/mv23682/Documents/Abil/studies/wiseman2024/env_data_processing/regridded_data/chlor_a.nc',
    '/home/mv23682/Documents/Abil/studies/wiseman2024/env_data_processing/regridded_data/Rrs_547.nc',
    '/home/mv23682/Documents/Abil/studies/wiseman2024/env_data_processing/regridded_data/Rrs_667.nc'    
]

# Open all datasets
datasets = [xr.open_dataset(fp) for fp in file_paths]

# Align all datasets to the intersection of their coordinates
aligned_datasets = xr.align(*datasets, join='inner')

# Merge the aligned datasets
merged_ds = xr.merge(aligned_datasets)

# Calculate CI_2
merged_ds['CI_2'] = merged_ds['Rrs_547'] - merged_ds['Rrs_667']

# Drop Rrs_547 and Rrs_667
merged_ds = merged_ds.drop_vars(['Rrs_547','Rrs_667'])

# List of variables of interest
variables_of_interest = ['temperature','sio4', 'po4', 'no3','o2','DIC','TA','PAR','mld','chlor_a','CI_2']  # Add all relevant variable names

# Create a mask for where any of the variables are NaN
variables_mask = xr.concat([merged_ds[var].isnull() for var in variables_of_interest], dim='var').any(dim='var')

# Create a mask that identifies where any of the variables are NaN across the depth dimension
#depth_nan_mask = xr.concat([merged_ds[var].isnull().any(dim='depth') for var in variables_of_interest], dim='var').any(dim='var')

# Expand the mask along the depth dimension to cover the entire column for each (lat, lon, time) slice
#depth_mask = depth_nan_mask.broadcast_like(merged_ds[variables_of_interest[0]])

# Combine both masks (any variable or depth missing) to set all data to NaN where either condition is true
#combined_mask = variables_mask | depth_mask
#combined_mask = depth_mask
combined_mask = variables_mask

# Apply the mask to all variables, setting values to NaN where any variable is missing
for var in variables_of_interest:
    merged_ds[var] = merged_ds[var].where(~combined_mask, np.nan)

# Convert the xarray Dataset to a Pandas DataFrame
df = merged_ds.to_dataframe()

# Reset index to flatten the multi-index DataFrame and drop NA
df = df.reset_index()
df.dropna(inplace=True)

# Save the DataFrame to a CSV file
df.to_csv('/home/mv23682/Documents/Abil/studies/wiseman2024/data/env_data.csv', index=False)

merged_ds['lat'].attrs['units'] = 'degrees_north'
merged_ds['lat'].attrs['long_name'] = 'latitude'

merged_ds['lon'].attrs['units'] = 'degrees_east'
merged_ds['lon'].attrs['long_name'] = 'longitude'

merged_ds['depth'].attrs['units'] = 'm'
merged_ds['depth'].attrs['positive'] = 'down'

merged_ds['time'].attrs['units'] = 'months'

merged_ds['temperature'].attrs['units'] = 'degrees_celsius'
merged_ds['temperature'].attrs['long_name'] = 'sea_water_temperature'
merged_ds['temperature'].attrs['description'] = 'Objectively analyzed mean fields for sea_water_temperature from WOA18 of Locarnini et al. (2019)'

merged_ds['sio4'].attrs['units'] = 'umol.kg-1'
merged_ds['sio4'].attrs['long_name'] = 'silicate'
merged_ds['sio4'].attrs['description'] = 'Objectively analyzed mean fields for moles_concentration_of_silicate_in_sea_water from WOA18 of Garcia et al. (2019)'

merged_ds['po4'].attrs['units'] = 'umol.kg-1'
merged_ds['po4'].attrs['long_name'] = 'phosphate'
merged_ds['po4'].attrs['description'] = 'Objectively analyzed mean fields for moles_concentration_of_phosphate_in_sea_water from WOA18 of Garcia et al. (2019)'

merged_ds['no3'].attrs['units'] = 'umol.kg-1'
merged_ds['no3'].attrs['long_name'] = 'nitrate'
merged_ds['no3'].attrs['description'] = 'Objectively analyzed mean fields for moles_concentration_of_nitrate_in_sea_water from WOA18 of Garcia et al. (2019)'

merged_ds['o2'].attrs['units'] = 'umol.kg-1'
merged_ds['o2'].attrs['long_name'] = 'dissolved oxygen'
merged_ds['o2'].attrs['description']= 'Objectively analyzed mean fields for mole_concentration_of_dissolved_molecular_oxygen_in_sea_water from WOA18 of Garcia et al. (2019)'

merged_ds['mld'].attrs['units'] = 'm'
merged_ds['mld'].attrs['long_name'] = 'Mixed Layer Depth (MLD) estimated through a threshold value of surface potential density equals to +0.03 kg/m3 with reference to the surface value taken at 10 m depth'
merged_ds['mld'].attrs['description'] = 'This MLD climatology is generated by monthly mapping (OI/kriging) of MLD estimations from individual ocean casts/profiles measuring pressure/depth, temperature and salinity, and coming from the ARGO program (argo.ucsd.edu ; argo.net), and from the NOAA NCEI World Ocean Database (www.ncei.noaa.gov/products/world-ocean-database): CTD, APB (sea mammals), OSD (bottles), MRB (moored buoys including e.g. TAO array), GLD (gliders), UOR (undulating ocean recorder, e.g. seasor), DRB (drifting buoys, including e.g. Ice Tethered Profilers). Original L3 file timestamp: Tue Jul 11 18:48:39 2023'

merged_ds['DIC'].attrs['units'] = 'umol.kg-1'
merged_ds['DIC'].attrs['long_name'] = 'dissolved inorganic carbon'
merged_ds['DIC'].attrs['description'] = 'Monthly climatology of total dissolved inorganic carbon (TCO2) centered in 1995 and obtained with NNGv2LDEO of Broullón et al. (2020)'

merged_ds['TA'].attrs['units'] = 'umol.kg-1'
merged_ds['TA'].attrs['long_name'] = 'total alkalinity'
merged_ds['TA'].attrs['description'] = 'Monthly climatology of total alkalinity obtained with NNGv2 of Broullón et al. (2019)'

merged_ds['PAR'].attrs['units'] = 'W.m-2'
merged_ds['PAR'].attrs['long_name'] = 'photosynthetically activate radiation'
merged_ds['PAR'].attrs['description'] = 'RS_PAR_ESM-based_fill_monthly_clim_1998-2022 from Castant et al. (2024)'

merged_ds['chlor_a'].attrs['units'] = 'mg.m-3'
merged_ds['chlor_a'].attrs['long_name'] = 'chlorophyll_a'
merged_ds['chlor_a'].attrs['description'] = 'Aqua MODIS Level 3 binned chlorophyll data monthly climatology, version R2022.0, 2002-07-01 to 2023-09-03 (NASA Ocean Biology Processing Group, 2022)'

merged_ds['CI_2'].attrs['units'] = 'sr-1'
merged_ds['CI_2'].attrs['long_name'] = 'Color Index 2'
merged_ds['CI_2'].attrs['description'] = 'Color Index 2 (CI2) from Mitchell et al. (2017) calculated using Aqua MODIS Level 3 binned Rrs_547 and Rrs_667 data monthly climatology, version R2022.0, 2002-07-01 to 2023-09-03 (NASA Ocean Biology Processing Group, 2022)'

# Save the result to a new NetCDF file
merged_ds.to_netcdf('/home/mv23682/Documents/Abil/studies/wiseman2024/data/env_data.nc')
print('fin')
# %%



# %%
# Regrid GOSML Mixed Layer Depth to 180x360x12 and repeat for depths 0-200m
from operator import index
import numpy as np
import xarray as xr
import xesmf as xe
import pandas as pd
path = "/home/mv23682/Documents/Abil/studies/wiseman2024/env_data_processing/raw_data/GOSML/mixed_layer_properties_mean.nc"


ds = xr.open_dataset(path)
ds_out = xr.Dataset({'lat': (['lat'], np.arange(-90, 90, 1)),
                     'lon': (['lon'], np.arange(-180, 180, 1))
                     })
print(ds)
regridder1 = xe.Regridder(ds, ds_out, method="bilinear", periodic=True)
dr_out = regridder1(ds['depth_mean'])
df = dr_out.to_dataframe(name="mld")
df.reset_index(inplace = True)
df['month'] = df['month'] + 0.5
df1 = df 
df2 = df
df1 = df1.assign(depth=0)
df2 = df2.assign(depth=205)
df = pd.concat([df1, df2])
df.rename(columns = {'latitude':'lat', 'longitude':'lon', 'month':'time'}, inplace = True)
df = df.groupby(['time', 'depth', 'lat', 'lon']).mean().reset_index()
df.set_index(['time', 'depth', 'lat', 'lon'], inplace=True)
print(df)
ds = df.to_xarray()
ds = ds.interp(depth=np.arange(0, 205, 5))
df = ds.to_dataframe()
df.reset_index(inplace = True)
df.set_index(['time', 'depth', 'lat', 'lon'], inplace=True)
ds = df.to_xarray()
ds.to_netcdf("/home/mv23682/Documents/Abil/studies/wiseman2024/env_data_processing/regridded_data/MLD.nc")

print("fin")
# %%



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'Abil'
copyright = '2025, Abil developers'
author = 'nanophyto'
release = '25.09.18'


# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
	'sphinx.ext.autodoc', 
	'sphinx.ext.coverage', 
	'sphinx.ext.napoleon',
    "sphinx_design",
    "sphinxcontrib.bibtex",
]

bibtex_bibfiles = ["library.bib"]
bibtex_default_style = "unsrt"     
bibtex_reference_style = "label" 

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']



# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'pydata_sphinx_theme'

html_theme_options = {
    "navbar_end": ["theme-switcher", "version-switcher"],  # Remove "this-page" from navbar
    "secondary_sidebar_items": []  # This removes the "This Page" section
}




from sklearn.model_selection import StratifiedKFold, KFold
import numpy as np

class UpsampledZeroStratifiedKFold:
    """
    Custom cross-validation generator with upsampling of zero instances for stratified folds.
    """

    def __init__(self, n_splits=3):
        """
        Initialize the stratified K-fold with upsampling.

        Parameters
        ----------
        n_splits : int, default=3
            Number of folds.
        """
        self.n_splits = n_splits

    def split(self, X, y, groups=None):
        """
        Generate train-test splits with upsampling of the minority class in the training data.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Feature matrix.
        y : array-like, shape (n_samples,)
            Target variable.
        groups : array-like, optional
            Group labels for the samples, used for group-based splitting. Not used in this method.

        Yields
        ------
        train_indices : np.ndarray
            Indices for the training set with upsampled minority class.
        test_indices : np.ndarray
            Indices for the test set.

        Notes
        -----
        - Converts `y` into a binary variable (`1` for non-zero values, `0` otherwise) for stratified sampling.
        - Upsamples the minority class in the training set to match the size of the majority class.
        - Uses `StratifiedKFold` for generating splits based on the binary target variable.
        """
        y_binary = np.where(y!=0, 1, 0)

        for rx, tx in StratifiedKFold(n_splits=self.n_splits).split(X,y_binary):
            nix = np.where(y_binary[rx]==0)[0]
            pix = np.where(y_binary[rx]==1)[0]
            pixu = np.random.choice(pix, size=nix.shape[0], replace=True)
            ix = np.append(nix, pixu)
            rxm = rx[ix]
            yield rxm, tx

    def get_n_splits(self, X, y, groups=None):
        return self.n_splits
    
    
class ZeroStratifiedKFold:
    """
    Custom cross-validation generator to handle zero-inflated targets with stratification.
    """

    def __init__(self, n_splits=3):
        """
        Initialize the stratified K-fold.

        Parameters
        ----------
        n_splits : int, default=3
            Number of folds.
        """
        self.n_splits = n_splits

    def split(self, X, y, groups=None):
        """
        Generate train-test splits with upsampling of the minority class in the training data.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Feature matrix.
        y : array-like, shape (n_samples,)
            Target variable.
        groups : array-like, optional
            Group labels for the samples, used for group-based splitting. Not used in this method.

        Yields
        ------
        train_indices : np.ndarray
            Indices for the training set with upsampled minority class.
        test_indices : np.ndarray
            Indices for the test set.

        Notes
        -----
        - Converts `y` into a binary variable (`1` for non-zero values, `0` otherwise) for stratified sampling.
        - Upsamples the minority class in the training set to match the size of the majority class.
        - Uses `StratifiedKFold` for generating splits based on the binary target variable.
        """
        y_binary = np.where(y!=0, 1, 0)

        # Check if there are any zeros in the array
        if any(element == 0 for element in y_binary):
            for rx, tx in StratifiedKFold(n_splits=self.n_splits).split(X,y_binary):
                yield rx, tx
        else:
            for rx, tx in KFold(n_splits=self.n_splits).split(X,y_binary):
                yield rx, tx

    def get_n_splits(self, X, y, groups=None):
        return self.n_splits




import numpy as np
from sklearn.compose import TransformedTargetRegressor
from sklearn.model_selection import GridSearchCV

class LogGridSearch:
    """
    Perform grid search with optional logarithmic transformation of the target variable.

    Supports evaluating models with no transformation, log-transformation, or both.
    """

    def __init__(self, m, verbose, cv, param_grid, scoring, regions=None):
        """
        Initialize the LogGridSearch.

        Parameters
        ----------
        m : estimator
            Base model for grid search.
        verbose : int
            Verbosity level of grid search.
        cv : int or cross-validation generator
            Cross-validation strategy.
        param_grid : dict
            Grid of parameters to search.
        scoring : str or callable
            Scoring metric.
        regions : optional
            Additional regions information (if required).
        """
        self.m = m
        self.verbose = verbose
        self.cv = cv
        self.param_grid = param_grid
        self.scoring = scoring
        self.regions = regions

    def do_log(self, x):
        """
        Apply natural logarithm transformation to the input values.

        Parameters
        ----------
        x : array-like
            Input values.

        Returns
        -------
        y : array-like
            Log-transformed values.
        """
        y = np.log(x+1)
        return(y)

    def do_nothing(self, x):
        """
        Apply no transformation to the input values.

        Parameters
        ----------
        x : array-like
            Input values.

        Returns
        -------
        y : array-like
            Non-transformed values.
        """
        return(x)

    def do_exp(self, x):
        """
        Apply exponential transformation to the input values.

        Parameters
        ----------
        x : array-like
            Input values.

        Returns
        -------
        y : array-like
            Exponentially transformed values.
        """
        y = np.exp(x)-1
        return(y)

    def transformed_fit(self, X, y, log, predictors):
        """
        Perform grid search with optional log transformation on the target variable.

        Parameters
        ----------
        X : pd.DataFrame or np.ndarray
            Feature matrix.
        y : pd.Series or np.ndarray
            Target variable.
        log : str
            Transformation mode: "yes" (log transform), "no" (no transform), or "both" (test both).
        predictors : list
            Predictors used for training (not directly used in the function).

        Returns
        -------
        GridSearchCV
            Fitted grid search instance for the best-performing model.

        Notes
        -----
        - Applies log transformation using `TransformedTargetRegressor` when `log="yes"`.
        - If `log="both"`, compares models with and without log transformation.
        - Uses `self.param_grid`, `self.scoring`, and `self.cv` for grid search.

        Raises
        ------
        ValueError
            If `log` is not "yes", "no", or "both".
        """
        # TODO: would encoding this as True/False/'both' be better?
        if log == "yes":

            model = TransformedTargetRegressor(self.m, func = self.do_log, inverse_func=self.do_exp)
            grid_search = GridSearchCV(model, param_grid = self.param_grid, scoring=self.scoring, refit=True,
                            cv = self.cv, verbose = self.verbose, return_train_score=True, error_score=-1e99)
            grid_search.fit(X, y)

        elif log == "no":

            model = TransformedTargetRegressor(self.m, func = self.do_nothing, inverse_func=self.do_nothing)
            grid_search = GridSearchCV(model, param_grid = self.param_grid, scoring=self.scoring, refit=True,
                            cv = self.cv, verbose = self.verbose, return_train_score=True, error_score=-1e99)
            grid_search.fit(X, y)

        elif log == "both":

            normal_m = TransformedTargetRegressor(self.m, func = self.do_nothing, inverse_func=self.do_nothing)
            grid_search1 = GridSearchCV(normal_m, param_grid = self.param_grid, scoring=self.scoring, refit=True,
                            cv = self.cv, verbose = self.verbose, return_train_score=True, error_score=-1e99)
            grid_search1.fit(X, y)

            log_m = TransformedTargetRegressor(self.m, func = self.do_log, inverse_func=self.do_exp)
            grid_search2 = GridSearchCV(log_m, param_grid = self.param_grid, scoring=self.scoring, refit=True,
                            cv = self.cv, verbose = self.verbose, return_train_score=True, error_score=-1e99)
            grid_search2.fit(X, y)

            if (grid_search1.best_score_ > grid_search2.best_score_):
                grid_search = grid_search1
                best_transformation = "nolog"

            elif (grid_search1.best_score_ < grid_search2.best_score_):
                grid_search = grid_search2
                best_transformation = "log"
            else:
                grid_search = grid_search1
                best_transformation = "nobest"
            grid_search.cv_results_['best_transformation'] = best_transformation

        return grid_search



# %%
# Regrid GOSML Mixed Layer Depth to 180x360x12 and repeat for depths 0-200m
from operator import index
import numpy as np
import xarray as xr
import xesmf as xe
import pandas as pd
path = "/home/mv23682/Documents/Abil/studies/wiseman2024/env_data_processing/raw_data/SEANOE/mld_dr003_ref10m_v2023.nc"
ds = xr.open_dataset(path)
ds = ds.drop_vars(["mask","n_obs"])
ds = ds.rename({"mld_dr003":"mld"})

# Convert the time coordinate from "days since 1-1-1 0:0:0" to datetime objects
time_dates = pd.to_datetime(ds['time'].values, unit='D', origin='1-1-1')

# Extract the month component from the datetime objects
months = time_dates.month

# Replace the time coordinate with the extracted months
ds = ds.assign_coords(time=("time", months))
ds_out = xr.Dataset({'lat': (['lat'], np.arange(-90, 90, 1)),
                     'lon': (['lon'], np.arange(-180, 180, 1))
                     })
regridder1 = xe.Regridder(ds, ds_out, method="conservative", periodic=True)
dr_out = regridder1(ds['mld'])
df = dr_out.to_dataframe(name="mld")
df.reset_index(inplace = True)
df1 = df 
df2 = df
df1 = df1.assign(depth=0)
df2 = df2.assign(depth=205)
df = pd.concat([df1, df2])
df.rename(columns = {'month':'time'}, inplace = True)
df = df.groupby(['time', 'depth', 'lat', 'lon']).mean().reset_index()
df.set_index(['time', 'depth', 'lat', 'lon'], inplace=True)
ds = df.to_xarray()
ds = ds.interp(depth=np.arange(0, 205, 5))
df = ds.to_dataframe()
df.reset_index(inplace = True)
df.set_index(['time', 'depth', 'lat', 'lon'], inplace=True)
ds = df.to_xarray()
ds.to_netcdf("/home/mv23682/Documents/Abil/studies/wiseman2024/env_data_processing/regridded_data/MLD_SEANOE.nc")

print("fin")
# %%



#%%
# Merge monthly WOA18 Silicate and subset to 0-200m

import numpy as np
import xarray as xr
import xesmf as xe
import pandas as pd

base_path = '/home/mv23682/Documents/Abil/studies/wiseman2024/env_data_processing/raw_data/WOA18/'
files = [f"{i:02}" for i in range(1, 13)]

sample_ds = xr.open_dataset(base_path + "woa18_all_i01_01.nc", decode_times=False,
                            drop_variables={'climatology_bounds','crs','depth_bnds',
                                            'i_dd','i_gp','i_ma','i_mn','i_oa','i_sd','i_se',
                                            'lat_bnds','lon_bnds','time'})

sample_df = sample_ds.to_dataframe().reset_index()
sample_df = sample_df[sample_df["depth"] <= 200]
sample_df = sample_df.set_index(['depth', 'lat', 'lon'])
sample_xr = sample_df.to_xarray().interp(depth=np.arange(0, 205, 5))
ds_out = xr.Dataset({'lat': (['lat'], np.arange(-90, 90, 1)),
                    'lon': (['lon'], np.arange(-180, 180, 1)),
                    'depth': (['depth'], np.arange(0, 205, 5))
                    })
regridder1 = xe.Regridder(sample_xr, ds_out, method="conservative", periodic=True)

ds_all = []
for i in range(0, 12):
    print(f"Processing month {i+1}...")
    ds = xr.open_dataset(base_path + "woa18_all_i" + files[i] + "_01.nc",
                         decode_times=False,drop_variables={'climatology_bounds','crs','depth_bnds',
                                                            'i_dd','i_gp','i_ma','i_mn','i_oa','i_sd','i_se',
                                                            'lat_bnds','lon_bnds','time'}
                         )
    df = ds.to_dataframe().reset_index()
    df = df[df["depth"] <= 200]
    df = df.set_index(['depth', 'lat', 'lon'])
    ds = df.to_xarray().interp(depth=np.arange(0, 205, 5))
    dr_out = regridder1(ds['i_an'],skipna=True, na_thres=0.75)
    df = dr_out.to_dataframe(name="sio4")
    df.reset_index(inplace = True)
    df['time'] = i+1
    df = df.groupby(['time', 'depth', 'lat', 'lon']).mean().reset_index()
    df.set_index(['time', 'depth', 'lat', 'lon'], inplace=True)
    ds = df.to_xarray()
    ds_all.append(ds)

ds = xr.merge(ds_all)
print("Regridding complete.")
df = ds.to_dataframe().reset_index()
df.set_index(['time', 'depth', 'lat', 'lon'], inplace=True)
ds = df.to_xarray().to_netcdf("/home/mv23682/Documents/Abil/studies/wiseman2024/env_data_processing/regridded_data/sio4.nc")

print("fin")

# %%



"""
Regressor Ensemble 
"""
#paths:
import os
#handling data:
import numpy as np
import pandas as pd
import xarray as xr
from yaml import load
from yaml import CLoader as Loader
#abil functions:
from abil.tune import ModelTuner as tune
from abil.predict import ModelPredictor as predict
from abil.post import AbilPostProcessor as post
#plotting:
import matplotlib.pyplot as plt
import seaborn as sns

os.chdir(os.path.join(".", "examples"))

#load configuration yaml:
with open('regressor_BATS.yml', 'r') as f:
    model_config = load(f, Loader=Loader)

#load example training data:
d = pd.read_csv(os.path.join("data", "bats_training.csv"))
#define target:
target= "Gephyrocapsa huxleyi"
#define predictors based on YAML:
predictors = model_config['predictors']
#split training data in X_train and y
y = d[target]
X_train = d[predictors]

#plot training data:
#first we setup a helper function to plot the rasters:
def plot_depth_time(ax, da, variable=None, vmin=None, vmax=None,
             cbar_label=None):
    if isinstance(da, pd.DataFrame):
        da = da.to_xarray()
    if isinstance(da, xr.Dataset):
        da = da[variable]  
    m=da.plot(ax=ax,add_colorbar=True,
              vmin=vmin,vmax=vmax)
    m.colorbar.set_label(cbar_label)
    return ax
# then we can make the plots
fig, ax = plt.subplots(figsize=(14,5))
d.set_index(['depth', 'time'], inplace=True) # for plotting
plot_depth_time(ax, d, variable=target, cbar_label=r'abundance (cells L$^{-1}$)')
ax.set_title("Gephyrocapsa huxleyi abundance at BATS")
plt.savefig('BATS_observations.png',dpi=300,bbox_inches='tight',facecolor='white')
plt.show()

#train your model:
m = tune(X_train, y, model_config)
m.train(model="rf")
m.train(model="xgb")
m.train(model="knn")

#load prediction data:
X_predict = pd.read_csv(os.path.join("data", "bats_prediction.csv"))
X_predict.set_index(['depth', 'time'], inplace=True)
X_predict = X_predict[predictors]

#plotting the prediction data
vars_list=[('temperature','Temperature','°C'),
           ('PAR', r'PAR ',r'W m$^{-2}$'),
           ('no3','Nitrate','µmol kg$^{-1}$'),
           ('TA','Total Alkalinity','µmol kg$^{-1}$')]
fig,axs=plt.subplots(4,1,figsize=(7,8),constrained_layout=True)
for ax,(v,title,unit),lab in zip(axs,vars_list,['A)','B)','C)', 'D)']):
    plot_depth_time(ax, X_predict, variable=v, vmin=None, vmax=None, cbar_label=unit)
    ax.set_title(rf'$\mathbf{{{lab}}}$ {title}',loc='left',fontsize=10,pad=6,y=1.02)
plt.savefig('BATS_environment.png',dpi=300,bbox_inches='tight',facecolor='white')
plt.show()

#predict your model:
m = predict(X_train, y, X_predict, model_config)
m.make_prediction()

# Posts
targets = np.array([target])
def do_post(statistic):
    m = post(X_train, y, X_predict, model_config, statistic, datatype="abundance")
    m.export_ds("BATS")

do_post(statistic="mean")
do_post(statistic="ci95_UL")
do_post(statistic="ci95_LL")

# Plot the results
ds = xr.open_dataset(os.path.join("ModelOutput", "regressor", "posts", "BATS_mean_abundance.nc"))
ds_LL = xr.open_dataset(os.path.join("ModelOutput", "regressor", "posts", "BATS_ci95_LL_abundance.nc"))
ds_UL = xr.open_dataset(os.path.join("ModelOutput", "regressor", "posts", "BATS_ci95_UL_abundance.nc"))

fig,axs=plt.subplots(3,1,figsize=(7,6),constrained_layout=True)
for ax,da,title,lab in [
    (axs[0], ds[target], 'Mean Abundance','A)'),
    (axs[1], ds_LL[target], '95% CI Lower Limit','B)'),
    (axs[2], ds_UL[target], '95% CI Upper Limit','C)')]:
    plot_depth_time(ax, da, vmin=0, cbar_label=r'abundance (cells L$^{-1}$)')
    ax.set_title(rf'$\mathbf{{{lab}}}$ {title}')
plt.savefig('BATS_predictions.png',dpi=300,bbox_inches='tight',facecolor='white')
plt.show()

# Check statistics
obs_abundance = pd.read_csv(os.path.join("ModelOutput", "regressor", "posts", 
                                "performance", "ens_performance.csv"))
# print
print(
    f"error metrics:\n"
    f"R2 = {obs_abundance['R2'][0]:.2f}\n"
    f"relative RMSE = {obs_abundance['rRMSE'][0]*100:.2f} (%)\n"
    f"relative MAE = {obs_abundance['rMAE'][0]*100:.2f} (%)"
)

# Merge obs and model for plotting
targets = np.array([target])
def do_stats(statistic):
    m = post(X_train, y, X_predict, model_config, statistic, datatype="abundance")
    m.merge_obs(file_name="BATS", index_cols=['depth', 'time']) #default is lat, lon, depth, time

do_stats(statistic="mean")

obs_abundance = pd.read_csv(os.path.join("ModelOutput", "regressor", "posts", 
                                "BATS_obs_abundance.csv"))

# plot
sns.regplot(data = obs_abundance, x = 'Gephyrocapsa huxleyi', 
           y= 'Gephyrocapsa huxleyi_mod')
plt.xlabel("obs")
plt.ylabel("model")
plt.title('Observed vs modelled abundances')
plt.savefig('BATS_obs_preds.png',dpi=300,bbox_inches='tight',facecolor='white')
plt.show()



"""
Gephyrocapsa huxleyi global distribution
"""
# load dependencies
import pandas as pd
import xarray as xr
import numpy as np
from yaml import load
from yaml import CLoader as Loader
import matplotlib.pyplot as plt
import cartopy.crs as ccrs
import cartopy.feature as cfeature
from abil.tune import tune
from abil.predict import predict
from abil.post import post
from abil.utils import upsample

import os
os.chdir('./paper')

# Load configuration yaml:
with open('./data/2-phase.yml', 'r') as f:
    model_config = load(f, Loader=Loader)

# Load training data:
d = pd.read_csv("./data/training.csv")
predictors = model_config['predictors']
target = "Gephyrocapsa huxleyi HET"
d = d.dropna(subset=predictors)
d[target] = d[target].fillna(0)
d = upsample(d, target, ratio=10)
y = d[target]
X_train = d[predictors]

# Train your model:
m = tune(X_train, y, model_config)
m.train(model="rf")
m.train(model="xgb")
m.train(model="knn")

# Predict your model:
X_predict = pd.read_csv("./data/env_mean_global_surface.csv")
X_predict.set_index(["lat", "lon"], inplace=True)
X_predict = X_predict[predictors]
X_predict = X_predict.dropna(subset=predictors)

m = predict(X_train, y, X_predict, model_config)
m.make_prediction()

# Posts
targets = np.array([target])
def do_post(statistic):
    m = post(X_train, y, X_predict, model_config, statistic, datatype="abundance")
    m.export_ds("my_first_2-phase_model")

do_post(statistic="mean")
do_post(statistic="ci95_UL")
do_post(statistic="ci95_LL")

# Load the data
ds = xr.open_dataset("./ModelOutput/ghux_example/posts/my_first_2-phase_model_mean_abundance.nc")
ds_UL = xr.open_dataset("./ModelOutput/ghux_example/posts/my_first_2-phase_model_ci95_UL_abundance.nc")
ds_LL = xr.open_dataset("./ModelOutput/ghux_example/posts/my_first_2-phase_model_ci95_LL_abundance.nc")
d = pd.read_csv("./data/training.csv")

# Log-transform data
ds['Gephyrocapsa huxleyi HET'] = np.log10(ds['Gephyrocapsa huxleyi HET'] + 1)
ds_UL['Gephyrocapsa huxleyi HET'] = np.log10(ds_UL['Gephyrocapsa huxleyi HET'] + 1)
ds_LL['Gephyrocapsa huxleyi HET'] = np.log10(ds_LL['Gephyrocapsa huxleyi HET'] + 1)
d['Gephyrocapsa huxleyi HET'] = np.log10(d['Gephyrocapsa huxleyi HET'])

# Create figure with adjusted subplot spacing
fig, axs = plt.subplots(2, 2, figsize=(12, 6),
                       subplot_kw={'projection': ccrs.PlateCarree()})
axs = axs.flatten()

# Titles and panel labels
titles = ['Training Data', 'Mean Abundance', '95% CI Lower Limit', '95% CI Upper Limit']
panel_labels = ['A)', 'B)', 'C)', 'D)']

# Custom title function to handle labels
def add_title(ax, title, label, y=1.1):
    ax.set_title(f'$\mathbf{{{label}}}$  {title}', loc='left', pad=10, y=y, fontsize=10)
    return ax

# Plot Training Data
sc = axs[0].scatter(d['lon'], d['lat'], c=d['Gephyrocapsa huxleyi HET'],
                   cmap='viridis', s=10, transform=ccrs.PlateCarree(),
                   vmin=0)
add_title(axs[0], titles[0], panel_labels[0])
axs[0].set_ylim([-90, 90])
cbar0 = plt.colorbar(sc, ax=axs[0], shrink=0.6, pad=0.1)
cbar0.ax.tick_params(labelsize=8)
cbar0.set_label('log$_{10}$ abundance (cells L$^{-1}$)', size=8)

# Plot Mean Abundance
p1 = ds['Gephyrocapsa huxleyi HET'].plot(ax=axs[1], cmap='viridis', add_colorbar=False,
                                    vmin=0)
add_title(axs[1], titles[1], panel_labels[1])
axs[1].set_ylim([-90, 90])
cbar1 = plt.colorbar(p1, ax=axs[1], shrink=0.6, pad=0.1)
cbar1.ax.tick_params(labelsize=8)
cbar1.set_label('log$_{10}$ abundance (cells L$^{-1}$)', size=8)

# Plot CI Lower Limit
p2 = ds_LL['Gephyrocapsa huxleyi HET'].plot(ax=axs[2], cmap='viridis', add_colorbar=False,
                                       vmin=0)
add_title(axs[2], titles[2], panel_labels[2])
axs[2].set_ylim([-90, 90])
cbar2 = plt.colorbar(p2, ax=axs[2], shrink=0.6, pad=0.1)
cbar2.ax.tick_params(labelsize=8)
cbar2.set_label('log$_{10}$ abundance (cells L$^{-1}$)', size=8)

# Plot CI Upper Limit
p3 = ds_UL['Gephyrocapsa huxleyi HET'].plot(ax=axs[3], cmap='viridis', add_colorbar=False,
                                       vmin=0)
add_title(axs[3], titles[3], panel_labels[3])
axs[3].set_ylim([-90, 90])
cbar3 = plt.colorbar(p3, ax=axs[3], shrink=0.6, pad=0.1)
cbar3.ax.tick_params(labelsize=8)
cbar3.set_label('log$_{10}$ abundance (cells L$^{-1}$)', size=8)

# Add coastlines
for ax in axs:
    ax.add_feature(cfeature.LAND, color='gray')
    ax.add_feature(cfeature.COASTLINE, linewidth=0.5)
    gl = ax.gridlines(draw_labels=True, linewidth=0.5, alpha=0.5)
    gl.xlabel_style = {'size': 8}
    gl.ylabel_style = {'size': 8}

plt.tight_layout()
plt.savefig('figure_1.png', dpi=300, bbox_inches='tight', facecolor='white')
plt.show()



# import required packages
import pandas as pd
import sys, os
from yaml import load
from yaml import CLoader as Loader
from abil.post import AbilPostProcessor
from datetime import datetime

current_date = datetime.today().strftime('%Y-%m-%d')

print(sys.argv[0])
dirpath = os.path.dirname(os.path.abspath(__file__))
conffile = os.path.abspath(os.path.join(dirpath,'ensemble_regressor.yml'))
root = os.path.abspath(os.path.join(dirpath,'..','..'))

#load model arguments
with open(conffile, 'r') as f:
    model_config = load(f, Loader=Loader)

#define model config:
model_config['hpc'] = True

#load data
targets = pd.read_csv(os.path.join(root,model_config['targets']))
target =  targets['Target'][0]
targets = targets['Target'].values
target_subset = targets[:2] # subset for estimate_applicability and merge_obs
d = pd.read_csv(os.path.join(root,model_config['training']))
predictors = model_config['predictors']
d = d.dropna(subset=predictors)

X_predict =  pd.read_csv(os.path.join(root,model_config['prediction']))
X_predict.set_index(['time','depth','lat','lon'],inplace=True)
X_predict = X_predict[model_config['predictors']]
X_predict = X_predict.dropna()

y = d[target]
X_train = d[predictors]

# Create a mask where y is a pseudo-zero
mask = y != 0
y = y[mask]
X_train = X_train[mask]

def do_post(statistic):
    m = AbilPostProcessor(X_train, y, X_predict, model_config, statistic)
    
    if statistic == "mean":
        print('begin estimate_applicability')
        m.estimate_applicability(target_subset)

        print('begin merge_obs')
        m.merge_obs(current_date,target_subset)
    
        print('begin process_resampled_runs')
        m.process_resampled_runs()

    print('begin export_ds')
    m.export_ds(current_date)

    print('begin integration')
    magnitude_conversion = 1e-21
    molar_mass = 12.01
    integ = m.integration(m, magnitude_conversion=magnitude_conversion,molar_mass=molar_mass,rate=True)
    if statistic == "mean":
        integ.integrated_totals(targets)
    else:
        integ.integrated_totals(target_subset)

    print('do_post for: ', statistic, ' complete')

do_post(statistic="mean")
do_post(statistic="ci95_UL")
do_post(statistic="ci95_LL")




# import required packages
import pandas as pd
import sys, os
from yaml import load
from yaml import CLoader as Loader
from abil.tune import ModelTuner

#define directories
print(sys.argv[1])
dirpath = os.path.dirname(os.path.abspath(__file__))
conffile = os.path.abspath(os.path.join(dirpath,'ensemble_regressor.yml'))
root = os.path.abspath(os.path.join(dirpath,'..','..'))

#load model arguments
with open(conffile, 'r') as f:
    model_config = load(f, Loader=Loader)

#define model config:
model_config['hpc'] = True
n_jobs = pd.to_numeric(sys.argv[1])
n_spp = pd.to_numeric(sys.argv[2])
model = sys.argv[3]
model_config['n_threads'] = n_jobs
model_config['cv'] = 10

#load data
targets = pd.read_csv(os.path.join(root,model_config['targets']))
d = pd.read_csv(os.path.join(root,model_config['training']))
target =  targets['Target'][n_spp]
predictors = model_config['predictors']
d = d.dropna(subset=[target])
d = d.dropna(subset=predictors)

y = d[target]
X_train = d[predictors]

#setup model:
m = ModelTuner(X_train, y, model_config)

#run model:
m.train(model=model, regressor=True)



import unittest
unittest.TestLoader.sortTestMethodsUsing = None

import sys, os
from yaml import load
from yaml import CLoader as Loader
import pandas as pd
import numpy as np
import xarray as xr 

from abil.tune import ModelTuner as tune
from abil.utils import example_data # example_training_data, example_predict_data
from abil.predict import ModelPredictor as predict
from abil.post import AbilPostProcessor as post

class TestRegressors(unittest.TestCase):

    def setUp(self):
        self.workspace = os.getenv('GITHUB_WORKSPACE', '.')
        with open(os.path.join(self.workspace,'tests/regressor.yml'), 'r') as f:
            self.model_config = load(f, Loader=Loader)

        self.model_config['local_root'] = self.workspace # yaml_path


        self.target_name =  "Emiliania huxleyi"
        self.X_train, self.X_predict, self.y = example_data(self.target_name, n_samples=1000, n_features=3, noise=0.1, train_to_predict_ratio=0.7, random_state=59)
#        self.X_predict = X_predict[predictors]


    def test_post_ensemble(self):
        m = tune(self.X_train, self.y, self.model_config)
        m.train(model="rf", log="yes")
        m.train(model="xgb", log="yes")
        m.train(model="knn", log="yes")

        m = predict(self.X_train, self.y, self.X_predict, self.model_config, n_jobs=self.model_config['n_threads'])
        m.make_prediction()

        # Load datasets
        rf = xr.open_dataset("./tests/ModelOutput/regressor/predictions/rf/Emiliania_huxleyi.nc")
        xgb = xr.open_dataset("./tests/ModelOutput/regressor/predictions/xgb/Emiliania_huxleyi.nc")  # Note: same path as rf?

        # Calculate sums
        rf_mean_sum = np.sum(rf['mean'])
        rf_ci95_UL_sum = np.sum(rf['ci95_UL'])
        xgb_mean_sum = np.sum(xgb['mean'])
        xgb_ci95_UL_sum = np.sum(xgb['ci95_UL'])

        print("======================")
        print("DEBUG OF PREDICT SUMS")
        print("======================")

        print(f"RF mean sum: {rf_mean_sum}")
        print(f"XGB mean sum: {xgb_mean_sum}")
        # Check if values are within same order of magnitude
#        if not np.isclose(rf_mean_sum, xgb_mean_sum, rtol=9):
#            raise AssertionError("Mean sums are not within an order of magnitude")
        
        print(f"RF ci95_UL sum: {rf_ci95_UL_sum}")
        print(f"XGB ci95_UL sum: {xgb_ci95_UL_sum}")
#        if not np.isclose(rf_ci95_UL_sum, xgb_ci95_UL_sum, rtol=9):
#            raise AssertionError("CI95 UL sums are not within an order of magnitude")

        targets = np.array([self.target_name])
        def do_post(statistic):
            m = post(self.X_train, self.y, self.X_predict, self.model_config, statistic, datatype="poc")
            #estimate aoa for each target and export to aoa.nc:
            m.estimate_applicability(drop_zeros=True)

            m.estimate_carbon("pg poc")

            m.total()

            m.merge_env()
            m.merge_obs("test",targets)

            m.export_ds("test")
            m.export_csv("test")

            vol_conversion = 1e3 #L-1 to m-3
            integ = m.integration(m, vol_conversion=vol_conversion)
            print(targets)
            integ.integrated_totals(targets)
            integ.integrated_totals(targets, monthly=True)

        do_post(statistic="mean")
        do_post(statistic="ci95_UL")
        do_post(statistic="ci95_LL")




class Test2Phase(unittest.TestCase):

    def setUp(self):
        self.workspace = os.getenv('GITHUB_WORKSPACE', '.')
        with open(os.path.join(self.workspace, 'tests/2-phase.yml'), 'r') as f:
            self.model_config = load(f, Loader=Loader)

        self.model_config['local_root'] = self.workspace # yaml_path


        self.target_name =  "Emiliania huxleyi"

        self.X_train, self.X_predict, self.y = example_data(self.target_name, n_samples=1000, n_features=3, noise=0.1, train_to_predict_ratio=0.7, random_state=59)
        

    def test_post_ensemble(self):


        m = tune(self.X_train, self.y, self.model_config)

        m.train(model="rf", log="yes")
        m.train(model="xgb", log="yes")
        m.train(model="knn", log="yes")

        m = predict(self.X_train, self.y, self.X_predict, self.model_config, n_jobs=self.model_config['n_threads'])
        m.make_prediction()
        print("======================")
        print("DEBUG OF PREDICT SUMS")
        print("======================")
        # Load datasets
        rf = xr.open_dataset("./tests/ModelOutput/2-phase/predictions/rf/Emiliania_huxleyi.nc")
        xgb = xr.open_dataset("./tests/ModelOutput/2-phase/predictions/xgb/Emiliania_huxleyi.nc")  # Note: same path as rf?

        # Calculate sums
        rf_mean_sum = np.sum(rf['mean'])
        rf_ci95_UL_sum = np.sum(rf['ci95_UL'])
        xgb_mean_sum = np.sum(xgb['mean'])
        xgb_ci95_UL_sum = np.sum(xgb['ci95_UL'])

        print(f"RF mean sum: {rf_mean_sum}")
        print(f"XGB mean sum: {xgb_mean_sum}")
        
        # Check if values are within same order of magnitude
#        if not np.isclose(rf_mean_sum, xgb_mean_sum, rtol=9):
#            raise AssertionError("Mean sums are not within an order of magnitude")
        
        print(f"RF ci95_UL sum: {rf_ci95_UL_sum}")
        print(f"XGB ci95_UL sum: {xgb_ci95_UL_sum}")
#        if not np.isclose(rf_ci95_UL_sum, xgb_ci95_UL_sum, rtol=9):
#            raise AssertionError("CI95 UL sums are not within an order of magnitude")

        targets = np.array([self.target_name])

        def do_post(statistic):
            m = post(self.X_train, self.y, self.X_predict, self.model_config, statistic, datatype="poc")
            #estimate aoa for each target and export to aoa.nc:
            m.estimate_applicability(drop_zeros=True)
            m.estimate_carbon("pg poc")
            m.diversity()

            m.total()
            m.merge_env()
            m.merge_obs("test",targets)

            m.export_ds("test")

            vol_conversion = 1e3 #L-1 to m-3
            integ = m.integration(m, vol_conversion=vol_conversion)
            integ.integrated_totals(targets, monthly=True)
            integ.integrated_totals(targets)

        do_post(statistic="mean")
        do_post(statistic="ci95_UL")
        do_post(statistic="ci95_LL")


def _lat_band_area(lat_deg, dlat_deg):
    """R^2 * (sin(phi+Δφ/2) - sin(phi-Δφ/2))  -> shape (nlat,)"""
    phi = np.deg2rad(np.asarray(lat_deg))
    dphi = np.deg2rad(float(dlat_deg))
    return (6_371_000.0**2) * (np.sin(phi + dphi/2.0) - np.sin(phi - dphi/2.0))

def _horiz_area_lat_lon(lat_deg, lon_deg, dlat_deg, dlon_deg):
    """Per-cell horizontal area for (lat, lon): A_latband(lat) * Δλ  -> (nlat, nlon)"""
    lat_band = _lat_band_area(lat_deg, dlat_deg)     # (nlat,)
    dlam = np.deg2rad(float(dlon_deg))
    lon_width = np.full(len(lon_deg), dlam)          # (nlon,)
    return np.outer(lat_band, lon_width)             # (nlat, nlon)

class _IntegrationParent:
    """Minimal parent to satisfy the integration class."""
    def __init__(self, ds: xr.Dataset):
        self.d = ds.to_dataframe()
        self.root = "."
        self.model_config = {"path_out": ".", "run_name": "testrun"}
        self.statistic = "stat"
        self.datatype = ""

class TestIntegrationFlexibleDims(unittest.TestCase):
    """Checks known totals for multiple dimension combos."""

    def setUp(self):
        # tiny grid & constants
        self.lat = np.array([-0.5, 0.5])       # 1° tall bands around equator
        self.lon = np.array([0.0, 1.0, 2.0])   # 1° wide cells
        self.depth = np.array([5.0, 15.0])     # two layers (indices only)
        self.time = np.array([1, 2], dtype=int)  # integer months (Jan=1, Feb=2)
        self.dlat = 1.0
        self.dlon = 1.0
        self.depth_w = 10.0  # m per depth cell
        self.value = 2.0     # uniform field

        # expected helpers
        self.area2d = _horiz_area_lat_lon(self.lat, self.lon, self.dlat, self.dlon)
        self.horiz_total = self.area2d.sum()
        self.depth_total = self.depth_w * len(self.depth)
        self.zonal_area_by_lat = (2.0 * np.pi) * _lat_band_area(self.lat, self.dlat)
        self.zonal_total = self.zonal_area_by_lat.sum()

    def _make_integration(self, ds, rate=False):
        return post.integration(
            parent=_IntegrationParent(ds),
            resolution_lat=self.dlat,
            resolution_lon=self.dlon,
            depth_w=self.depth_w,
            rate=rate
        )

    def test_lat_lon_depth_time(self):
        coords = {"lat": self.lat, "lon": self.lon, "depth": self.depth, "time": self.time}
        data = np.full((len(self.lat), len(self.lon), len(self.depth), len(self.time)), self.value)
        ds = xr.Dataset({"foo": xr.DataArray(data, dims=("lat","lon","depth","time"), coords=coords)})

        integ = self._make_integration(ds, rate=False)
        total = integ.integrate_total(variable="foo", monthly=False)

        expected = self.value * self.horiz_total * self.depth_total * len(self.time)
        self.assertAlmostEqual(float(total.values), expected, places=10)

    def test_lat_lon_only(self):
        coords = {"lat": self.lat, "lon": self.lon}
        data = np.full((len(self.lat), len(self.lon)), self.value)
        ds = xr.Dataset({"foo": xr.DataArray(data, dims=("lat","lon"), coords=coords)})

        integ = self._make_integration(ds, rate=False)
        total = integ.integrate_total(variable="foo", monthly=False)

        expected = self.value * self.horiz_total
        self.assertAlmostEqual(float(total.values), expected, places=10)

    def test_lat_depth_only(self):
        coords = {"lat": self.lat, "depth": self.depth}
        data = np.full((len(self.lat), len(self.depth)), self.value)
        ds = xr.Dataset({"foo": xr.DataArray(data, dims=("lat","depth"), coords=coords)})

        integ = self._make_integration(ds, rate=False)
        total = integ.integrate_total(variable="foo", monthly=False)

        expected = self.value * self.zonal_total * self.depth_total
        self.assertAlmostEqual(float(total.values), expected, places=10)

    def test_time_only(self):
        coords = {"time": self.time}
        data = np.full((len(self.time),), self.value)
        ds = xr.Dataset({"foo": xr.DataArray(data, dims=("time",), coords=coords)})

        integ = self._make_integration(ds, rate=False)
        total = integ.integrate_total(variable="foo", monthly=False)

        expected = self.value * len(self.time)  # volume=1, just sum over time
        self.assertAlmostEqual(float(total.values), expected, places=10)


if __name__ == '__main__':
    # Create a test suite combining all test cases in order
    suite = unittest.TestSuite()
    suite.addTest(TestRegressors('test_post_ensemble'))
    suite.addTest(Test2Phase('test_post_ensemble'))

    # post area integration tests
    suite.addTest(TestIntegrationFlexibleDims('test_lat_lon_depth_time'))
    suite.addTest(TestIntegrationFlexibleDims('test_lat_lon_only'))
    suite.addTest(TestIntegrationFlexibleDims('test_lat_depth_only'))
    suite.addTest(TestIntegrationFlexibleDims('test_time_only'))

    runner = unittest.TextTestRunner()
    runner.run(suite)



import numpy as np
import pandas as pd
import logging
logger = logging.getLogger("abil")

from sklearn.base import BaseEstimator, RegressorMixin, clone, is_regressor, is_classifier

class ZeroInflatedRegressor(BaseEstimator, RegressorMixin):
    """
    A custom regressor to handle zero-inflated target variables.

    Combines a classifier to predict non-zero occurrences and a regressor for non-zero targets.
    """

    def __init__(self, classifier, regressor, threshold=0.5):
        """
        Initialize the regressor with a classifier and regressor.

        Parameters
        ----------
        classifier : estimator
            A classifier to predict non-zero values.
        regressor : estimator
            A regressor to predict non-zero targets.
        threshold : float
            The probability cutoff for predicting presence 

        """
        self.classifier = classifier
        self.regressor = regressor
        self.threshold = threshold

    def fit(self, X, y, sample_weight=None):
        """
        Fit the model.

        Parameters
        ----------
        X : np.ndarray of shape (n_samples, n_features)
            The training data.

        y : np.ndarray, 1-dimensional
            The target values.

        sample_weight : Optional[np.array], default=None
            Individual weights for each sample.

        Returns
        -------
        ZeroInflatedRegressor
            Fitted regressor.

        Raises
        ------
        ValueError
            If `classifier` is not a classifier or `regressor` is not a regressor.
        """
        if not is_classifier(self.classifier):
            raise ValueError(
                f"`classifier` has to be a classifier. Received instance of {type(self.classifier)} instead.")
        if not is_regressor(self.regressor):
            raise ValueError(f"`regressor` has to be a regressor. Received instance of {type(self.regressor)} instead.")

        # Ensure classifier_ is assigned
        self.classifier_ = clone(self.classifier)
        self.classifier_.fit(X, y != 0)

        # Ensure regressor_ is assigned
        self.regressor_ = clone(self.regressor)
        
        y_pred_proba = self.classifier_.predict_proba(X)[:, 1]
        y_pred = (y_pred_proba >= self.threshold).astype(int)
        non_zero_indices = np.where(y_pred == 1)[0]

        if non_zero_indices.size > 0:
            if isinstance(X, pd.DataFrame):
                self.regressor_.fit(
                    X.iloc[non_zero_indices] if isinstance(X, pd.DataFrame) else X[non_zero_indices],
                    y.iloc[non_zero_indices].values if isinstance(y, pd.Series) else y[non_zero_indices]
                )
            else:
                self.regressor_.fit(
                        X[non_zero_indices],
                        y[non_zero_indices],
                )
        else:
            logger.info("All predictions are zero (!), skipping regressor fitting.")
        
        return self


    def predict(self, X):
        """
        Get predictions.

        Parameters
        ----------
        X : np.ndarray, shape (n_samples, n_features)
            Samples to get predictions of.

        Returns
        -------
        y : np.ndarray, shape (n_samples,)
            The predicted values.
        """
        output = np.zeros(len(X))

        non_zero_indices = np.where(self.classifier_.predict(X) == 1)[0]

        if non_zero_indices.size > 0:
            if isinstance(X, pd.DataFrame):
                output[non_zero_indices] = self.regressor_.predict(X.iloc[non_zero_indices])
            else:
                output[non_zero_indices] = self.regressor_.predict(X[non_zero_indices])

        return output



"""
2-phase Ensemble 
"""
#paths:
import os
#handling data:
import numpy as np
import pandas as pd
import xarray as xr
from yaml import load
from yaml import CLoader as Loader
#abil functions:
from abil.tune import ModelTuner as tune
from abil.predict import ModelPredictor as predict
from abil.post import AbilPostProcessor as post
from abil.utils import example_data 
#plotting:
import matplotlib.pyplot as plt
import cartopy.crs as ccrs
import cartopy.feature as cfeature
from matplotlib.colors import Normalize

os.chdir(os.path.join(".", "examples"))

#load configuration yaml:
with open('2-phase.yml', 'r') as f:
    model_config = load(f, Loader=Loader)

#load example training data:
d = pd.read_csv(os.path.join("data", "so_training.csv"))
#define target:
target = "Gephyrocapsa huxleyi HET"
#define predictors based on YAML:
predictors = model_config['predictors']
#split training data in X_train and y
y = d[target]
X_train = d[predictors]

#train your model:
m = tune(X_train, y, model_config)
m.train(model="rf")
m.train(model="xgb")
m.train(model="knn")

#load prediction data:
X_predict = pd.read_csv(os.path.join("data", "so_prediction.csv"))
X_predict.set_index(['lat', 'lon'], inplace=True)
X_predict = X_predict[predictors]

#predict your model:
m = predict(X_train, y, X_predict, model_config)
m.make_prediction()

# Posts
targets = np.array([target])
def do_post(statistic):
    m = post(X_train, y, X_predict, model_config, statistic, datatype="abundance")
    m.export_ds("SO") #Southern Ocean

do_post(statistic="mean")
do_post(statistic="ci95_UL")
do_post(statistic="ci95_LL")

# Load the predictions
ds = xr.open_dataset(os.path.join("ModelOutput", "2-phase", "posts", "SO_mean_abundance.nc"))
ds_LL = xr.open_dataset(os.path.join("ModelOutput", "2-phase", "posts", "SO_ci95_LL_abundance.nc"))
ds_UL = xr.open_dataset(os.path.join("ModelOutput", "2-phase", "posts", "SO_ci95_UL_abundance.nc"))

# Create the figure
def plot_panel(ax, data, var, title, label, 
               cbar_label='abundance (cells L$^{-1}$)'):
    ax.set_extent([-180,180,-90,-30], crs=ccrs.PlateCarree())
    ax.coastlines(); ax.gridlines(draw_labels=False, linewidth=0.5, alpha=0.5)
    ax.add_feature(cfeature.LAND, color='gray') 
    ax.add_feature(cfeature.COASTLINE, linewidth=0.5)
    ax.set_title(f'$\mathbf{{{label}}}$  {title}', loc='left', 
                 pad=10, y=1.05, fontsize=10)

    if isinstance(data, pd.DataFrame):  # raw points
        lon, lat = np.asarray(data['lon']), np.asarray(data['lat'])
        vals = np.asarray(data[var])
        # clip outliers for plotting
        vmin, vmax = np.nanpercentile(vals, [2,98])
        x, y = ax.projection.transform_points(ccrs.PlateCarree(), lon, lat)[:,0:2].T
        h = ax.hexbin(x, y, C=vals, reduce_C_function=np.nanmean, gridsize=20, 
                      mincnt=1, norm=Normalize(vmin=vmin, vmax=vmax))
    else:  # gridded data
        da = data[var] if isinstance(data, xr.Dataset) else data
        h = da.plot(ax=ax, add_colorbar=False, robust=True, 
                    transform=ccrs.PlateCarree())

    cb = plt.colorbar(h, ax=ax, shrink=0.6, pad=0.1); cb.ax.tick_params(labelsize=8)
    cb.set_label(cbar_label, size=8)

# apply plotting function and export
fig, axs = plt.subplots(2,2, figsize=(8,6), 
                        subplot_kw={'projection': ccrs.SouthPolarStereo()})
(ax00, ax01), (ax10, ax11) = axs
plot_panel(ax00, d, target, 'Training Data', 'A)')
plot_panel(ax01, ds,    target, 'Mean Abundance', 'B)')
plot_panel(ax10, ds_LL, target, '95% CI Lower Limit', 'C)')
plot_panel(ax11, ds_UL, target, '95% CI Upper Limit', 'D)')
plt.tight_layout()
plt.savefig('SO_predictions.png', dpi=300, bbox_inches='tight', facecolor='white')
plt.show()

# we can also look at integrated totals (stocks)
def do_integrations(statistic):
    m = post(X_train, y, X_predict, model_config, statistic, datatype="pic")
    m.estimate_carbon("pg pic")
    vol_conversion = 1e3 # to convert from pg C L-1 to pg C m-3
    magnitude_conversion = 1e-24 # convert from pg C to Tg C
    vol_conversion = vol_conversion*200 # model approximates the top 200m
    integ = m.integration(m, vol_conversion=vol_conversion,
                          magnitude_conversion=magnitude_conversion)
    integ.integrated_totals(targets)

do_integrations(statistic="mean")
do_integrations(statistic="ci95_UL")
do_integrations(statistic="ci95_LL")

mean = pd.read_csv(os.path.join("ModelOutput", "2-phase", "posts", 
                                "integrated_totals", 
                                "ens_integrated_totals_mean_pic.csv"))['total'][0]
ci95_LL = pd.read_csv(os.path.join("ModelOutput", "2-phase", "posts", 
                                "integrated_totals", 
                                "ens_integrated_totals_ci95_LL_pic.csv"))['total'][0]
ci95_UL = pd.read_csv(os.path.join("ModelOutput", "2-phase", "posts", 
                                "integrated_totals", 
                                "ens_integrated_totals_ci95_UL_pic.csv"))['total'][0]

print(f"estimated integrated total: {mean:.2f} [{ci95_LL:.2f}, {ci95_UL:.2f}] Tg IC")



import numpy as np
import xarray as xr

# Load the NetCDF file
file_path = '/home/mv23682/Documents/Abil/studies/wiseman2024/env_data_processing/raw_data/RS_PAR/RS_PAR_ESM-based_fill_monthly_clim_1998-2022.nc'
ds = xr.open_dataset(file_path, mode="r+")

# Rename coordinates
ds = ds.rename({'latitude':'lat','longitude':'lon','Depth':'depth','Date':'time','Time':'time'})

# Save the changes
ds.to_netcdf('/home/mv23682/Documents/Abil/studies/wiseman2024/env_data_processing/regridded_data/par.nc')

# Close the dataset
ds.close()


import numpy as np
import pandas as pd

from sklearn.metrics import mean_absolute_error, balanced_accuracy_score

from sklearn.model_selection import KFold
from sklearn.preprocessing import FunctionTransformer
from sklearn.ensemble import (
    RandomForestRegressor,
    RandomForestClassifier,
    BaggingRegressor,
    BaggingClassifier,
    VotingRegressor,
    VotingClassifier,
)
from sklearn.compose import TransformedTargetRegressor
from sklearn.neighbors import KNeighborsRegressor, KNeighborsClassifier
from sklearn.pipeline import Pipeline
from sklearn import base
from joblib import delayed, Parallel

from sklearn.base import is_regressor

from .zir import ZeroInflatedRegressor
from . import utils as u
from sklearn.base import _is_fitted
from xgboost import XGBClassifier, XGBRegressor, DMatrix


def estimate_prediction_quantiles(
    model, X_predict, X_train, y_train, cv=None, chunksize=20_000, threshold=0.5
):
    """
    Train the model using cross-validation, compute predictions on X_train with summary stats,
    and predict on X_predict with summary stats.

    Parameters:
    -----------
    X_train : DataFrame
        Training feature set with MultiIndex for coordinates.

    y_train : Series
        Target values corresponding to X_train.

    X_predict : DataFrame
        Feature set to predict on, with MultiIndex for coordinates.

    m : sklearn pipeline

    cv_splits : int, default=5
        Number of cross-validation splits.

    method : str, default="rf"
        Method type for handling different model-specific behaviors:
        "rf" for RandomForestRegressor,
        "bagging" for BaggingRegressor,
        "xgb" for XGBRegressor.

    Returns:
    --------
    dict
        Dictionary containing summary statistics for both training and prediction datasets.
        Keys: "train_stats", "predict_stats".
    """

    #########################################
    # IF ZIR ESTIMATE QUANTILES RECURSIVELY #
    #########################################

    if isinstance(model, ZeroInflatedRegressor):
        original_y_train = y_train.copy()

        classifier_stats = estimate_prediction_quantiles(
            model.classifier_,
            X_predict = X_predict,
            X_train = X_train,
            y_train = (y_train > 0).copy(),
            cv=cv,
            chunksize=chunksize,
            threshold=threshold
        )
        regressor_stats = estimate_prediction_quantiles(
            model.regressor_,
            X_predict=X_predict,
            X_train=X_train,
            y_train=original_y_train,
            cv=cv,
            chunksize=chunksize,
            threshold=threshold
        )
        return {
            **{f"classifier_{k}": v for k, v in classifier_stats.items()},
            **{f"regressor_{k}": v for k, v in regressor_stats.items()},
        }

    #######################################
    # EXTRACT BASE MODEL AND TRANSFORMERS #
    #######################################

    if isinstance(model, Pipeline):
        pipeline = model
        model = pipeline.named_steps["estimator"]
        y_inverse_transformer = u.do_nothing
        y_transformer = u.do_nothing
    elif isinstance(model, TransformedTargetRegressor):
        pipeline = model.regressor
        y_transformer = getattr(
            model, "func", FunctionTransformer().func
        )

        y_inverse_transformer = getattr(
            model, "inverse_func", FunctionTransformer().inverse_func
        )

        model = model.regressor.named_steps["estimator"]
    else:
        pipeline = Pipeline(
            [("preprocessor", FunctionTransformer()), ("estimator", model)]
        )
        y_inverse_transformer = u.do_nothing
        y_transformer = u.do_nothing

    preprocessor = pipeline.named_steps["preprocessor"]
    preprocessor.fit(X_train)

    ##################################################
    # CHECK IF BASE MODEL IS REGRESSOR OR CLASSIFIER #
    ##################################################

    if is_regressor(model):
        proba = False
    else:
        proba = True


    ############################################
    # TRANSFORM Y_TRAIN, X_TRAIN AND X_PREDICT #
    ############################################

    if X_train is not None:
        X_train = pd.DataFrame(
            preprocessor.transform(X_train),
            index = getattr(X_train, "index", np.arange(X_train.shape[0])),
            columns=X_train.columns
        )
    else:
        raise ValueError("X_train cannot be None")
    if X_predict is not None:
        X_predict = pd.DataFrame(
            preprocessor.transform(X_predict),
            index = getattr(X_predict, "index", np.arange(X_predict.shape[0])),
            columns=X_predict.columns
        )
    else:
        raise ValueError("X_predict cannot be None")

    if y_train is not None:
        if not proba:
            y_train = pd.Series(
                y_transformer(y_train),
                index = getattr(y_train, "index", np.arange(y_train.shape[0]))
            )

        else:
            y_train = pd.Series(
                y_train,
                index = getattr(y_train, "index", np.arange(y_train.shape[0]))
            )
    else:
        raise ValueError("y_train cannot be None")

    ##########################################
    # REFIT MODELS WITH TRANSFORMED Y AND Xs #
    ##########################################

    if not _is_fitted(model):
        model.fit(X_train, y_train)


    ###############################################
    # MAKE QUANTILE PREDICTIONS FOR TRAINING DATA #
    ###############################################

    if cv is not None:
        train_summary_stats = [
            _summarize_predictions(
                model,
                y_inverse_transformer,
                X_predict=X_train.iloc[test_idx],
                X_train=X_train.iloc[train_idx],
                y_train=y_train.iloc[train_idx],
                chunksize=chunksize,
                threshold=threshold,
                proba=proba
            )
            for train_idx, test_idx in cv.split(X_train, y_train)
        ]
        # concat and rearrange to match input data order
        train_summary_stats = pd.concat(
            train_summary_stats, axis=0, ignore_index=False
        ).loc[X_train.index]
    else:
        train_summary_stats = _summarize_predictions(
            model,
            y_inverse_transformer,
            X_train=X_train,
            X_predict=X_train,
            y_train=y_train,
            chunksize=chunksize,
            threshold=threshold,
            proba=proba
        )

    #################################################
    # MAKE QUANTILE PREDICTIONS FOR PREDICTION DATA #
    #################################################

    predict_summary_stats = _summarize_predictions(
        model,
        y_inverse_transformer,
        X_predict=X_predict,
        X_train=X_train,
        y_train=y_train,
        chunksize=chunksize,
        threshold=threshold,
        proba=proba
    )

    return {"train_stats": train_summary_stats, "predict_stats": predict_summary_stats}

#######################################
# DEFINE QUANTILE PREDICTION PIPELINE #
#######################################

def _summarize_predictions(model, y_inverse_transformer, X_predict, X_train=None, y_train=None, chunksize=2e4, threshold=0.5, proba=None):
    if proba is None:
        raise ValueError("proba not defined!")
    n_samples, n_features = X_predict.shape

    ######################################
    # DEFINE CHUNKS (FOR X_PREDICT ONLY) #
    ######################################

    if chunksize is not None:
        n_chunks = int(np.ceil(n_samples / chunksize))
        chunks = np.array_split(X_predict, n_chunks)
        # Convert each chunk to pandas DataFrame
        chunks = [pd.DataFrame(chunk, columns=X_predict.columns) for chunk in chunks]
    else:
        chunks = [X_predict]

    stats = []

    engine = Parallel()

    ##############################################################
    # ESTIMATE LOSS FOR EACH MEMBER BASED ON X_TRAIN PREDICTIONS #
    ##############################################################

    if X_train is not None and y_train is not None:
        if isinstance(model, (XGBRegressor)):
            model = BaggingRegressor(estimator=model, n_estimators=160)
            model.fit(X_train, y_train)
        elif isinstance(model, (XGBClassifier)):
            model = BaggingClassifier(estimator=model, n_estimators=160)
            model.fit(X_train, y_train)
        train_pred_jobs = _setup_pred_jobs(
            model,
            X_train,
            proba,
            threshold
        )
        train_results = engine(train_pred_jobs)

        if proba:
            losses = 1 - np.array(
                [balanced_accuracy_score(y_train.astype(int),
                pred.astype(int)) for pred in train_results]
            )

        else:
            losses = np.array([mean_absolute_error(y_train, y_inverse_transformer(pred)) for pred in train_results])

        weights = 1 / (losses + 1e-99)  # Avoid division by zero
        weights /= weights.sum()  # Normalize weights
    else:
        raise ValueError("X_train and y_train should be defined")

    ##########################################
    # COMPUTE PREDICTIONS BASED ON X_PREDICT #
    ##########################################

    for chunk in chunks:
        predict_pred_jobs = _setup_pred_jobs(model=model,
                                             X=chunk,
                                             proba=proba,
                                             threshold=threshold)

        results = engine(predict_pred_jobs)
        chunk_preds = pd.DataFrame(
            y_inverse_transformer(np.column_stack(results)),
            index=getattr(chunk, "index", None),
        )

        weights = None
        if weights is not None:
            lower = chunk_preds.apply(u.weighted_quantile, q=0.025, weights=weights, axis=1)
            upper =  chunk_preds.apply(u.weighted_quantile, q=0.975, weights=weights, axis=1)
        else:
            lower = chunk_preds.quantile(q=0.025, axis=1)
            upper =  chunk_preds.quantile(q=0.975, axis=1)

        chunk_stats = pd.DataFrame(
            np.column_stack((lower, upper)), # stats we've calculated by chunk
            columns=['ci95_LL', 'ci95_UL'], # column names for the stats
            index = chunk.index # the index of the chunk, so we can align the results back to X_predict
        )
        stats.append(chunk_stats)

    output = pd.concat(stats, axis=0, ignore_index=False)
    return output

##################################
# SETUP PARALLEL PREDICTION JOBS #
##################################

def _setup_pred_jobs(model, X, proba, threshold):
    if isinstance(model, (XGBClassifier, XGBRegressor)):
        n_estimators = u.xgboost_get_n_estimators(model)
        if n_estimators==None:
            raise ValueError("n_estimators not inferred correctly!")
        booster = model.get_booster()
        pred_jobs = (
            delayed(u._predict_one_member)(
                i,
                member=booster,
                chunk=X,
                proba=proba,
                threshold=threshold
            )
            for i in range(n_estimators)
        )
    else:
        members, features_for_members = _flatten_metaensemble(model)
        pred_jobs = (
            delayed(u._predict_one_member)(
                _,
                member=member,
                chunk=X.iloc[:, features_for_member],
                proba=proba,
                threshold=threshold
            )
            for _, (features_for_member, member) in enumerate(zip(features_for_members, members))
        )
    return pred_jobs


#################################################################
# EXTRACT INDIVIDUAL MEMBERS FROM SKLEARN ENSEMBLE-BASED MODELS #
#################################################################

def _flatten_metaensemble(me):
    """Modified version to handle RandomForest's individual trees"""
    estimators = []
    features = []

    # Case 1: RandomForest - get all its trees
    if hasattr(me, 'estimators_') and isinstance(me, (RandomForestRegressor, RandomForestClassifier)):
        n_features = me.n_features_in_
        estimators = me.estimators_
        features = [list(range(n_features))] * len(estimators)  # All trees use all features (but possibly different subsets)

    # Case 2: Bagging estimator (like BaggingRegressor with KNN)
    elif hasattr(me, 'estimators_') and hasattr(me, 'estimators_features_'):
        estimators = me.estimators_
        features = me.estimators_features_

    # Case 3: Regular ensemble (Voting, Stacking)
    elif hasattr(me, 'estimators_'):
        n_features = me.n_features_in_
        estimators = me.estimators_
        features = [list(range(n_features))] * len(estimators)

    # Case 4: Single estimator
    else:
        estimators = [me]
        features = [list(range(getattr(me, 'n_features_in_', X_train.shape[1])))]  # Fallback to X_train shape

    return estimators, features


# Example Usage
if __name__ == "__main__":
    # Generate sample data
    from sklearn.datasets import make_regression
    from joblib import parallel_backend  # this is user-facing
    from abil.utils import ZeroInflatedRegressor, ZeroStratifiedKFold

    X, y = make_regression(n_samples=100, n_features=10, noise=0.1)
    X_train = pd.DataFrame(X, columns=[f"feature_{i}" for i in range(X.shape[1])])
    latitudes = np.random.uniform(-90, 90, size=X_train.shape[0])
    longitudes = np.random.uniform(-180, 180, size=X_train.shape[0])
    X_train.index = pd.MultiIndex.from_tuples(
        zip(latitudes, longitudes), names=["Latitude", "Longitude"]
    )
    y_train = pd.Series(y)

    X_predict, _ = make_regression(n_samples=20, n_features=10, noise=0.1)
    X_predict = pd.DataFrame(
        X_predict, columns=[f"feature_{i}" for i in range(X_predict.shape[1])]
    )
    latitudes_predict = np.random.uniform(-90, 90, size=X_predict.shape[0])
    longitudes_predict = np.random.uniform(-180, 180, size=X_predict.shape[0])
    X_predict.index = pd.MultiIndex.from_tuples(
        zip(latitudes_predict, longitudes_predict), names=["Latitude", "Longitude"]
    )

    cv_splits = 5
    # Define cross-validation strategy
    cv = KFold(n_splits=cv_splits)

    # Define model and method
    model = RandomForestRegressor(n_estimators=100, max_depth=4, random_state=42).fit(
        X_train, y_train
    )

    vmodel = VotingRegressor(
        estimators=[
            (
                "rf",
                RandomForestRegressor(n_estimators=100, max_depth=4, random_state=42),
            ),
            (
                "knn",
                BaggingRegressor(
                    KNeighborsRegressor(n_neighbors=10, weights="distance"),
                    n_estimators=50,
                ).fit(X_train, y_train),
            ),
        ]
    ).fit(X_train, y_train)

    mask = (y_train > 0).values
    zirmodel = ZeroInflatedRegressor(
        BaggingClassifier(
            KNeighborsClassifier(n_neighbors=10, weights="distance"), n_estimators=50
        ),        BaggingRegressor(

            KNeighborsRegressor(n_neighbors=10, weights="distance"), n_estimators=50
        ),
    ).fit(X_train, y_train)

    v_of_zirmodels = VotingRegressor(
        estimators=[
            ("rf-then-knn", zirmodel),
            (
                "knn-then-rf",
                ZeroInflatedRegressor(
                    BaggingClassifier(
                        KNeighborsClassifier(n_neighbors=10, weights="distance"),
                        n_estimators=50,
                    ).fit(X_train, mask),
                    RandomForestRegressor(
                        n_estimators=100, max_depth=4, random_state=2245
                    ).fit(X_train.iloc[mask, :], y_train.iloc[mask]),
                ),
            ),
        ]
    ).fit(X_train, y_train)

    zir_of_vmodels = ZeroInflatedRegressor(
        classifier=VotingClassifier(
            estimators=[
                ("rfc", RandomForestClassifier(
                    n_estimators=100, max_depth=4, random_state=2245
                )),
                ("bagnnc", BaggingClassifier(
                    KNeighborsClassifier(n_neighbors=10, weights="distance"),
                    n_estimators=50,
                )),
            ]
        ),
        regressor=VotingRegressor(
            estimators=[
                ("rfr", RandomForestRegressor(n_estimators=100, max_depth=4, random_state=2245)),
                ("bagnnr", BaggingRegressor(
                    KNeighborsRegressor(n_neighbors=10, weights="distance"),
                    n_estimators=50,
                ))
            ]
        ),
    ).fit(X_train, y_train)

    # this sets the backend type and number of jobs to use in the internal
    # Parallel() call.
    with parallel_backend("loky", n_jobs=16):
        results = estimate_prediction_quantiles(
            model, X_predict=X_predict, X_train=X_train, y_train=y_train, cv=cv
        )
    with parallel_backend("loky", n_jobs=16):
        vresults = estimate_prediction_quantiles(
            vmodel, X_predict=X_predict, X_train=X_train, y_train=y_train, cv=cv
        )
    with parallel_backend("loky", n_jobs=16):
        zirresults = estimate_prediction_quantiles(
            zirmodel, X_predict=X_predict, X_train=X_train, y_train=y_train, cv=cv
        )
    with parallel_backend("loky", n_jobs=16):
        vzir_results = estimate_prediction_quantiles(
            zir_of_vmodels, X_predict=X_predict, X_train=X_train, y_train=y_train, cv=cv
        )
    logging.info(f"\n=== Training Summary Stats ===\n{results['train_stats'].head()}")
    logging.info(f"\n=== Prediction Summary Stats ===\n{results['predict_stats'].head()}")

    logging.info(f"\n=== Training Summary Stats ===\n{vzir_results['classifier_train_stats'].head()}")
    logging.info(f"\n=== Prediction Summary Stats ===\n{vzir_results['classifier_predict_stats'].head()}")

    logging.info(f"\n=== Training Summary Stats ===\n{vzir_results['regressor_train_stats'].head()}")
    logging.info(f"\n=== Prediction Summary Stats ===\{vzir_results['regressor_predict_stats'].head()}")



from sklearn import metrics
from sklearn import inspection
from sklearn import base
import numpy


def area_of_applicability(
    X_test,
    X_train,
    y_train=None,
    model=None,
    cv=None,
    metric="euclidean",
    feature_weights="permutation",
    feature_weight_kwargs=None,
    threshold="tukey",
    return_all=False,
    chunk_size=10000
):
    """
    Estimate the area of applicability for the data using a strategy similar to Meyer & Pebesma 2022).

    This calculates the importance-weighted feature distances from test to train points,
    and then defines the "applicable" test sites as those closer than some threshold
    distance.

    Parameters
    ----------
    X_test  :   numpy.ndarray
        array of features to be used in the estimation of the area of applicability
    X_train :   numpy.ndarray
        the training features used to calibrate cutoffs for the area of applicability
    y_train :   numpy.ndarray
        the outcome values to estimate feature importance weights. Must be provided
        if the permutation feature importance is calculated.
    model   :   sklearn.BaseEstimator
        the model for which the feature importance will be calculated. Must be provided
        if the permutation feature importance is calculated.
    cv      : sklearn.BaseCrossValidator
        the crossvalidator to use on the input training data, in order to calculate the 
        within-sample extrapolation threshold
    metric  :   str (Default: 'euclidean')
        the name of the metric used to calculate feature-based distances.
    feature_weights : str or numpy.ndarray (Default: 'permutation')
        the name of the feature importance weighting strategy to be used. By default,
        scikit-learn's permutation feature importance is used. Pre-calculated
        feature importance scores can also be used. To ignore feature importance,
        set feature_weights=False.
    feature_weight_kwargs : dict()
        options to pass to the feature weight estimation function. By default, these
        are passed directly to sklearn.inspection.permutation_importance()
    threshold   :   str or float (Default: 'tukey')
        how to calculate the cutoff value to determine whether a model is applicable
        for a given test point. This cutoff is calculated within the training
        data, and applied to the test data.
        - 'tukey': use the tukey rule, setting the cutoff at 1.5 times the inter-quartile range (IQR) above the upper hinge (75th percentile) for the train data
        dissimilarity index.
        - 'mad': use a median absolute deviation rule, setting the cutoff at three times
        the median absolute deviation above the median train data dissimilarity index
        - float: if a value between zero and one is provided, then the cutoff is set at the
        percentile provided for the train data dissimilarity index.
    return_all: bool (Default: False)
        whether to return the dissimilarity index and density of train points near the test 
        point. Specifically, the dissimilarity index is the distance from test to train points in feature space, divided by the average distance between training points. The local density is the count of training datapoints whose feature distance is closer than the threshold value.

    Returns
    -------
        If return_local_density=False, the output is a numpy.ndarray of shape (n_training_samples, ) describing where a model
        might be considered "applicable" among the test samples.

        If return_local_density=True, then the output is a tuple of numpy arrays.
        The first element is the applicability mentioned above, the second is the 
        dissimilarity index for the test points, and the thord
        is the local density of training points near each test point.

        A value of 0 indicates the point is within the Area of Applicability, 
        while a value of 1 indicates the point is outside the Area of Applicability.
    """
    if feature_weight_kwargs is None:
        feature_weight_kwargs = dict()
    
    base.check_array(X_test, force_all_finite='allow-nan')
    base.check_array(X_train, force_all_finite='allow-nan')

    mask_test = numpy.asarray(~numpy.isnan(X_test).any(axis=1))
    mask_train = numpy.asarray(~numpy.isnan(X_train).any(axis=1))
    
    if y_train is not None:
        mask_train *= (~numpy.isnan(numpy.asarray(y_train)))

    n_test, n_features = X_test.shape
    n_train, _ = X_train.shape

    if n_features != X_train.shape[1]:
        raise ValueError(
        "features must be the same for both training and test data."
    )

    if mask_test.sum() <= 2:
        raise ValueError("at least two test samples must be not nan")
    if mask_train.sum() <= 2:
        raise ValueError("at least two train samples must be not nan")

    aoa_ = numpy.empty((n_test, ))
    di_test_ = numpy.empty((n_test, ))
    lpd_test_ = numpy.empty((n_test, ))

    if return_all:
        lpd_test_ = numpy.empty((n_test, ))
        test_to_train_d_ = numpy.empty((n_test, n_train))

    X_train = X_train[mask_train]
    y_train = y_train[mask_train]

    if not feature_weights:
        feature_weights = numpy.ones(n_features)
    elif feature_weights == "permutation":
        if model is None:
            raise ValueError(
                "Model must be provided if permutation feature importance is used"
            )
        feature_weight_kwargs.setdefault("n_jobs", -1)
        feature_weight_kwargs.setdefault("n_repeats", 10)
        feature_weights = inspection.permutation_importance(
            model, X_train, y_train, **feature_weight_kwargs
        ).importances_mean
        feature_weights /= feature_weights.sum()
    else:
        if len(feature_weights) != n_features:
            raise ValueError(
            "weights must be provided for all features"
        )
        feature_weights = feature_weights / sum(feature_weights)

    train_distance = metrics.pairwise_distances(
        X_train * feature_weights[None, :], metric=metric
    )
    numpy.fill_diagonal(train_distance, train_distance.max())

    if cv is not None:
        d_mins = numpy.empty((X_train.shape[0],))
        mean_acc_num = 0
        mean_acc_den = 0
        for test_ix, train_ix in cv.split(X_train):
            hold_to_seen_d = train_distance[test_ix.reshape(-1,1), train_ix]
            d_mins[test_ix] = hold_to_seen_d.min(axis=1)
            mean_acc_num += hold_to_seen_d.sum()
            mean_acc_den += hold_to_seen_d.size
        d_mean = mean_acc_num/mean_acc_den    
    else:
        d_mins = train_distance.min(axis=1)
        numpy.fill_diagonal(train_distance, 0)
        d_mean = d_mean = train_distance[train_distance > 0].mean()

    di_train = d_mins / d_mean

    if threshold == "tukey":
        lo_hinge, hi_hinge = numpy.percentile(di_train, (0.25, 0.75))
        iqr = hi_hinge - lo_hinge
        cutpoint = iqr * 1.5 + hi_hinge
    elif threshold == "mad":
        median = numpy.median(di_train)
        mad = numpy.median(numpy.abs(di_train - median))
        cutpoint = median + 3 * mad
    elif (0 < threshold) & (threshold < 1):
        cutpoint = numpy.percentile(di_train, threshold)
    cutpoint = numpy.maximum(cutpoint, di_train.max())

    if return_all:
        test_to_train_d = metrics.pairwise_distances(
            X_test[mask_test] * feature_weights[None, :],
            # NOTE: X_train is already filtered from the previous step!
            X_train * feature_weights[None, :],
            metric=metric,
        )  

        # NOTE: This is gross, but I don't see how to work around this. 
        # I tried doing something along the lines of: 
        # ```{python}
        # test_to_train_d_[
        #     mask_test.reshape(-1,1)*mask_test.reshape(1,-1)
        # ] = test_to_train_d
        # ```
        # but you can't assign a 2d matrix *into* a 2d selection. 
        test_to_train_d_[:] = numpy.nan
        i_in_full = 0
        for i, is_null in enumerate(mask_test):
            if not is_null:
                test_to_train_d_[i, mask_train] = test_to_train_d[i_in_full]
                i_in_full+=1

        test_to_train_d_min = test_to_train_d.min(axis=1)
        test_to_train_i = test_to_train_d.argmin(axis=1)

        di_test = test_to_train_d_min / d_mean
        lpd_test = ((test_to_train_d / test_to_train_d.mean()) < cutpoint).sum(axis=1)

    else:
        # if we don't need local point density, this can be used
        test_to_train_i, test_to_train_d_min = metrics.pairwise_distances_argmin_min(
            X_test[mask_test] * feature_weights[None, :],
            # NOTE: X_train is already filtered from the previous step!
            X_train * feature_weights[None, :],
            metric=metric,
        )
        di_test = test_to_train_d_min / d_mean
        lpd_test = numpy.empty_like(di_test) * numpy.nan

    aoa = di_test >= cutpoint

    lpd_test_[mask_test] = lpd_test
    di_test_[mask_test] = di_test
    aoa_[mask_test] = aoa
    lpd_test_[~mask_test] = di_test_[~mask_test] = aoa_[~mask_test] = numpy.nan
    
    if return_all:
        return aoa_, di_test_, lpd_test_, cutpoint, test_to_train_d_
    else:
        return aoa_, di_test_, cutpoint

if __name__ == "__main__":
    import pandas as pd
    import numpy as np
    from sklearn.datasets import make_regression
    from sklearn.model_selection import KFold
    from sklearn.neighbors import KNeighborsRegressor
    from sklearn.model_selection import GridSearchCV

    # Generate X and y
    X, y = make_regression(n_samples=200, n_features=10, noise=0.1, random_state=42)
    X = pd.DataFrame(X, columns=[f"feature_{i}" for i in range(X.shape[1])])

    # Add random latitude and longitude and set as MultiIndex
    latitudes = np.random.uniform(-90, 90, size=X.shape[0])
    longitudes = np.random.uniform(-180, 180, size=X.shape[0])
    X.index = pd.MultiIndex.from_tuples(zip(latitudes, longitudes), names=['Latitude', 'Longitude'])

    # The full X is the same as X_predict
    X_predict = X.copy()

    # Take a random subset and call this X_train and y_train
    sample_indices = np.random.choice(range(len(X)), size=100, replace=False) 
    X_train = X.iloc[sample_indices] 
    y_train = y[sample_indices]  

    # Initialize and fit the model
    reg = GridSearchCV(
        estimator = KNeighborsRegressor(),
        param_grid={"n_neighbors": [5, 8]},
        cv = KFold(n_splits=5))

    reg.fit(X_train, y_train)
    model = reg.best_estimator_   #this is what is currently loaded into post

    # Convert y_train to a pandas Series
    # in the pipeline y_train = df['species_name'], where df is a multi-index pandas dataframe
    y_train = pd.Series(y_train, index=X_train.index)

    aoa = area_of_applicability(
        X_test=X_predict,
        X_train=X_train,
        y_train=y_train,
        model=model
    )
    aoa2 = area_of_applicability(
        X_test=X_predict + 3,
        X_train=X_train,
        y_train=y_train,
        model=model,
        cv=reg.cv
    )
    ynan_train = y_train.copy()
    train_nans = X_train.sample(frac=.2).sample(frac=.2, axis=1)
    Xnan_train = X_train.copy()
    Xnan_train.loc[train_nans.index, train_nans.columns] = numpy.nan
    ynan_train[~ynan_train.index.isin(train_nans.index)].iloc[0] = numpy.nan
    Xnan_predict = X_predict.copy()
    test_nans = Xnan_predict.sample(frac=.2).sample(frac=.2,axis=1)
    Xnan_predict.loc[test_nans.index, test_nans.columns] = numpy.nan


    nantest = area_of_applicability(
        X_test=Xnan_predict,
        X_train=Xnan_train,
        y_train=ynan_train,
        model=model,
        return_all=True
    )

    if not numpy.isnan(
        nantest[0][X_predict.index.isin(test_nans.index)]
    ).all():
        raise ValueError("nan check failed, some nan values were dropped!")
    



import time
import os
import pickle
import pandas as pd
import numpy as np
import logging
logger = logging.getLogger("abil")

from joblib import parallel_backend
from xgboost import XGBClassifier, XGBRegressor
from sklearn.model_selection import GridSearchCV, cross_validate, KFold
from sklearn.ensemble import RandomForestRegressor, RandomForestClassifier, BaggingRegressor, BaggingClassifier
from sklearn.neighbors import KNeighborsRegressor, KNeighborsClassifier
from sklearn.pipeline import Pipeline
from sklearn.compose import ColumnTransformer
from sklearn.preprocessing import OneHotEncoder,  StandardScaler
from sklearn.gaussian_process import GaussianProcessClassifier, GaussianProcessRegressor

from .zir import ZeroInflatedRegressor
from .zero_stratified_kfold import ZeroStratifiedKFold,  UpsampledZeroStratifiedKFold
from .log_grid_search import LogGridSearch

class ModelTuner:
    """
    A class for model training, hyperparameter tuning, and cross-validation.

    Attributes
    ----------
    X_train : pd.DataFrame
        The feature matrix used for training the models.
    y : pd.Series
        The target variable.
    model_config : dict
        Configuration dictionary containing model and training parameters.
    regions : str or None, optional
        Name of the feature column representing regions, used for stratification (default is None).

    Methods
    -------
    train(model, classifier=False, regressor=False, log="no"):
        Train and tune models based on the provided configuration.
    """

    def __init__(self, X_train, y, model_config, regions=None):
        """
        Initialize the `ModelTuner` object.

        Parameters
        ----------
        X_train : pd.DataFrame of shape (n_samples, n_features)
            Training features used for model fitting.
        y : pd.Series of shape (n_samples,) or (n_samples, n_outputs)
            Target values used for model fitting.
        model_config : dict
            Dictionary containing model configuration parameters such as:
                - seed: int, random seed for reproducibility
                - root : str, path to Abil root folder
                - path_out : str, where predictions are saved
                - target : str, file name of your target list
                - verbose : int, to set verbosity (0-3)
                - n_threads : int, number of threads to use
                - cv : int, number of cross-folds
                - ensemble_config : dict
                    Dictionary containing ensemble set up:
                        - classifier: bool
                            Whether to train a classification model.
                        - regressor: bool
                            Whether to train a regression model.
                        - m{n}: str, model name (ex. m1: "rf", m2: "xgb" etc.)
        regions : str or None, optional
            Column name for regions to be used in preprocessing and stratification.

        Examples
        --------
        >>> from sklearn.ensemble import RandomForestClassifier
        >>> from sklearn.datasets import make_classification
        >>> with open('/home/phyto/Abil/configuration/example_model_config.yml', 'r') as f:
        ...    model_config = load(f, Loader=Loader)
        >>> X, y = example_data(y_name =  "Coccolithus pelagicus",
        ...                            n_samples=500, n_features=5, noise=20, 
        ...                            random_state=model_config['seed'])
        >>> m = ModelTuner(X, y, model_config)

        Returns
        -------
        m: object
            The model used for training.
        
        """
        self.y = y.sample(frac=1, random_state=model_config['seed']) #shuffle
        logger.info(f"length of y: {len(self.y)}")
        self.y = self.y.values.ravel()
        self.X_train = X_train.sample(frac=1, random_state=model_config['seed']) #shuffle
        self.model_config = model_config
        self.ensemble_config = model_config['ensemble_config']
        self.seed = model_config['seed']
        self.target = y.name
        self.target_no_space = self.target.replace(' ', '_')
        self.n_jobs = model_config['n_threads']
        self.verbose = model_config['verbose'] 
        self.regions = regions
        self.path_out = os.path.join(model_config['root'], model_config['path_out'], model_config['run_name'])

        # Check for valid regions
        if regions is not None:
            if regions not in X_train.columns:
                raise ValueError("Regions defined but not in X_train. Did you mean regions=None?")

        # Setup cross-validation strategy
        if model_config['stratify']:
            if model_config['upsample']:
                self.cv = UpsampledZeroStratifiedKFold(n_splits=model_config['cv'])
                logger.info("upsampling = True")
            else:
                self.cv = ZeroStratifiedKFold(n_splits=model_config['cv'])
        else:
            self.cv = KFold(n_splits=model_config['cv'])
             
        self.bagging_estimators = model_config.get('knn_bagging_estimators', None)

        # Preprocessor for features
        predictors = model_config['predictors'].copy()
        numeric_features = X_train.columns.get_indexer(X_train[predictors].columns)
        numeric_transformer = Pipeline(steps=[('scaler', StandardScaler())])

        if regions:
            model_config['predictors'].remove(regions)
            categorical_features = [self.regions]
            categorical_transformer = OneHotEncoder(handle_unknown='ignore')
            self.preprocessor = ColumnTransformer(transformers=[
                ('num', numeric_transformer, numeric_features),
                ('cat', categorical_transformer, categorical_features)
            ])
        else:
            self.preprocessor = ColumnTransformer(transformers=[
                ('num', numeric_transformer, numeric_features)
            ])
            

    
    def train(self, model, classifier=False, regressor=False, log="no"):

        """
        Trains a machine learning model using the specified configuration.

        Parameters
        ----------
        model : str
            The type of model to train. Supported options:
            - 'rf' : Random Forest
            - 'knn' : K-Nearest Neighbors
            - 'xgb' : XGBoost
            - 'gp' : Gaussian Process
        classifier : bool, default=False
            Whether to train a classification model.
        regressor : bool, default=False
            Whether to train a regression model.
        log : str, default="no"
            Log transformation option:
            - 'yes' : Apply log transformation to the target variable.
            - 'no' : No transformation.
            - 'both' : Train both with and without log transformation.
        
        Examples
        --------
        >>> m.train(model="rf", regressor=True)

        Returns
        -------
        None
        """

        if model =="xgb":
            clf_estimator = XGBClassifier(nthread=1, random_state=self.seed)
            reg_estimator = XGBRegressor(nthread=1, random_state=self.seed, 
                                         objective='reg:tweedie')
        elif model=="knn":
            if self.bagging_estimators ==None:
                raise ValueError("number of bagging estimators not defined")
            else:
                clf_estimator = BaggingClassifier(estimator=KNeighborsClassifier(), 
                                                  n_estimators=self.bagging_estimators, 
                                                  random_state=self.seed)
                reg_estimator = BaggingRegressor(estimator=KNeighborsRegressor(), 
                                                 n_estimators=self.bagging_estimators, 
                                                 random_state=self.seed)
        elif model=="rf":
            clf_estimator = RandomForestClassifier(random_state=self.seed, oob_score=True)
            reg_estimator = RandomForestRegressor(random_state=self.seed, oob_score=True)
        elif model=="gp":
            from sklearn.gaussian_process.kernels import RBF

            kernel = 1 * RBF(length_scale=2.0, length_scale_bounds=(1e-2, 1e2))

            clf_estimator = GaussianProcessClassifier(random_state=self.seed)
            reg_estimator =  GaussianProcessRegressor(random_state=self.seed, kernel=kernel,
                                                       normalize_y=True)
        else:
            raise ValueError("invalid model")

        if (not self.ensemble_config['classifier']) and (not self.ensemble_config['regressor']):
            raise ValueError("both classifier and regressor defined as False")

        if self.ensemble_config['classifier'] and (not self.ensemble_config['regressor']):        
            raise ValueError("classifiers are not supported")

        if self.ensemble_config['regressor']:
            if self.ensemble_config['classifier']:
                y = self.y[self.y > 0]
                X_train = self.X_train[self.y > 0].reset_index(drop=True)
                cv = ZeroStratifiedKFold(n_splits=self.model_config['cv'])
            else:
                y = self.y
                X_train = self.X_train
                cv = self.cv

            logger.info("training regressor")

            reg_scoring = {
                'R2': 'r2',
                'MAE': 'neg_mean_absolute_error',
                'RMSE': 'neg_root_mean_squared_error'
            }

            user_reg_param_grid = self.model_config['param_grid'][model + '_param_grid']['reg_param_grid']

            # Add the prefix 'regressor__estimator__' to each key
            reg_param_grid = {
                f"regressor__estimator__{key}": value
                for key, value in user_reg_param_grid.items()
            }

            logger.info(reg_param_grid)

            reg_sav_out_scores = os.path.join(self.path_out, "scoring/", model)
            reg_sav_out_model = os.path.join(self.path_out, "model/", model)


            os.makedirs(reg_sav_out_scores, exist_ok=True)
            os.makedirs(reg_sav_out_model, exist_ok=True)         
                
            reg_pipe = Pipeline(steps=[('preprocessor', self.preprocessor),
                        ('estimator', reg_estimator)])
            
            with parallel_backend('loky', n_jobs=self.n_jobs):
                reg = LogGridSearch(reg_pipe, verbose = self.verbose, cv=cv, 
                                    param_grid=reg_param_grid, scoring='r2', regions=self.regions)
                reg_grid_search = reg.transformed_fit(X_train, y, log, self.model_config['predictors'].copy())

            m2 = reg_grid_search.best_estimator_


            with open(os.path.join(reg_sav_out_model, self.target_no_space) + '_reg.sav', 'wb') as f:
                pickle.dump(m2, f)

            logger.info(f"exported model to: {reg_sav_out_model + '/'  + self.target_no_space + '_reg.sav'}")

            with parallel_backend('loky', n_jobs=self.n_jobs):
                reg_scores = cross_validate(m2, X_train, y, cv = cv, verbose = self.verbose, scoring=reg_scoring)

            with open(os.path.join(reg_sav_out_scores, self.target_no_space) + '_reg.sav', 'wb') as f:
                pickle.dump(reg_scores, f)

            logger.info(f"exported scoring to:  {reg_sav_out_scores + '/' + self.target_no_space + '_reg.sav'}")

            if "RMSE" in reg_scoring:
                try:
                    logger.info(f"reg rRMSE: {np.mean(reg_scores['test_RMSE'])/np.mean(self.y)*-100:.0f}%")
                except ValueError:
                    logger.info("reg rRMSE is NA (!)")
            if "MAE" in reg_scoring:
                try:
                    logger.info(f"reg rMAE: {np.mean(reg_scores['test_MAE'])/np.mean(self.y)*-100:.0f}%")
                except ValueError:
                    logger.info("reg rMAE is NA (!)")
            if "R2" in reg_scoring:
                try:
                    logger.info(f"reg R2: {np.mean(reg_scores['test_R2']):.2f}")
                except ValueError:
                    logger.info("reg R2 is NA (!)")

        if self.ensemble_config['classifier'] and self.ensemble_config['regressor']:      
            
            logger.info("training classifier")

            user_clf_param_grid = self.model_config['param_grid'][model + '_param_grid']['clf_param_grid']

            # Add the prefix 'regressor__estimator__' to each key
            clf_param_grid = {
                f"estimator__{key}": value
                for key, value in user_clf_param_grid.items()
            }
            
            clf_scoring = {'accuracy': 'balanced_accuracy'}

            clf_sav_out_scores = os.path.join(self.path_out, "scoring/", model)
            clf_sav_out_model = os.path.join(self.path_out, "model/", model)


            
            os.makedirs(clf_sav_out_scores, exist_ok=True)
            os.makedirs(clf_sav_out_model, exist_ok=True)


            clf_pipe = Pipeline(steps=[('preprocessor', self.preprocessor),
                      ('estimator', clf_estimator)])

            clf = GridSearchCV(
                estimator=clf_pipe,
                param_grid= clf_param_grid,
                scoring= 'balanced_accuracy',
                cv = self.cv,
                verbose = self.verbose
            )

            y_clf =  self.y.copy()
            logger.info(f"length of y_clf: {len(y_clf)}")

            y_clf[y_clf > 0] = 1
            logger.info(y_clf)
            with parallel_backend('loky', self.n_jobs):
                clf.fit(self.X_train, y_clf)

            m1 = clf.best_estimator_
            
            with open(os.path.join(clf_sav_out_model, self.target_no_space) + '_clf.sav', 'wb') as f:
                pickle.dump(m1, f)
            
            logger.info(f"exported model to: {clf_sav_out_model + '/' + self.target_no_space + '_clf.sav'}")

            clf_scores = cross_validate(m1, self.X_train, y_clf, cv=self.cv, verbose =self.verbose, scoring=clf_scoring)
            
            with open(os.path.join(clf_sav_out_scores, self.target_no_space) + '_clf.sav', 'wb') as f:
                pickle.dump(clf_scores, f)
            
            logger.info(f"exported scoring to: {clf_sav_out_scores + '/' + self.target_no_space + '_clf.sav'}")

            logger.info(clf_scores['test_accuracy'])
            logger.info(f"clf balanced accuracy {np.mean(clf_scores['test_accuracy']):.2f}")

            logger.info("training zero-inflated regressor")

            zir = ZeroInflatedRegressor(
                classifier=m1,
                regressor=m2,
            )

            zir_sav_out_scores = os.path.join(self.path_out, "scoring/", model)
            zir_sav_out_model = os.path.join(self.path_out, "model/", model)


            os.makedirs(zir_sav_out_scores, exist_ok=True)
            os.makedirs(zir_sav_out_model, exist_ok=True)

            zir.fit(self.X_train, self.y)

            with open(os.path.join(zir_sav_out_model, self.target_no_space) + '_zir.sav', 'wb') as f:
                pickle.dump(zir, f)
                
            logger.info(f"exported model to: {zir_sav_out_model + '/' + self.target_no_space + '_zir.sav'}")

            with parallel_backend('loky', n_jobs=self.n_jobs):
                zir_scores = cross_validate(zir, self.X_train, self.y, cv=self.cv, verbose =self.verbose, scoring=reg_scoring)

            with open(os.path.join(zir_sav_out_scores, self.target_no_space) + '_zir.sav', 'wb') as f:
                pickle.dump(zir_scores, f)

            logger.info(f"exported scoring to: {zir_sav_out_scores + '/' + self.target_no_space + '_zir.sav'}")

            try:
                logger.info(f"zir rRMSE: {(np.mean(zir_scores['test_RMSE'])/np.mean(self.y)*-100):.0f}%")
            except ValueError:
                logger.info("zir rRMSE is NA (!)")
            try:
                logger.info(f"zir rMAE: {(np.mean(zir_scores['test_MAE'])/np.mean(self.y)*-100):.0f}%")
            except ValueError:
                logger.info("zir rMAE is NA (!)")
            try:
                logger.info(f"zir R2: {np.mean(zir_scores['test_R2']):.2f}")
            except ValueError:
                logger.info("zir R2 is NA (!)")
        st = time.time()
        et = time.time()
        elapsed_time = et-st

        logger.info(f"execution time: {elapsed_time:.1f} seconds")        


if __name__ == "__main__":
    import os
    os.environ['MKL_NUM_THREADS'] = '1'
    os.environ['OMP_NUM_THREADS'] = '1'
    os.environ['MKL_DYNAMIC'] = 'FALSE'
    os.environ["OMP_THREAD_LIMIT"] = "1"




import pandas as pd
import numpy as np
import pickle
import os
import time
import logging
logger = logging.getLogger("abil")

from sklearn.ensemble import VotingRegressor, VotingClassifier
from sklearn.model_selection import KFold, cross_validate
from joblib import Parallel, delayed, parallel_backend  

from .utils import inverse_weighting, find_optimal_threshold
import shutil

# Set the custom temporary folder for loky
temp_folder = os.path.join(".","tmp") 
os.environ["LOKY_TEMP_FOLDER"] = temp_folder
# Ensure the directory exists
os.makedirs(temp_folder, exist_ok=True)

from . import unified_tree_or_bag as pp
from .zir import ZeroInflatedRegressor
from .zero_stratified_kfold import ZeroStratifiedKFold,  UpsampledZeroStratifiedKFold

def load_model_and_scores(path_out, ensemble_config, n, target):
    """
    Loads a trained model and scoring information, and calculates the mean absolute error (MAE) for the prediction.

    Parameters
    ----------
    path_out : str
        Path to the output folder containing model and scoring information.
    ensemble_config : dict
        Dictionary containing configuration details for the ensemble models, including model names, regressor or classifier status.
    n : int
        Index of the model to load in the ensemble.
    target : str
        The target for which predictions are made (used to load target-specific files).

    Returns
    -------
    tuple
        A tuple containing the model and the mean absolute error (MAE) score.

    Raises
    ------
    ValueError
        If both regressor and classifier are set to False, or if a classifier is used when only regressors are supported.
    """

    path_to_scores  = os.path.join(path_out, "scoring", ensemble_config["m"+str(n+1)])
    path_to_param  = os.path.join(path_out, "model", ensemble_config["m"+str(n+1)])

    if (ensemble_config["classifier"] ==True) and (ensemble_config["regressor"] == False):
        raise ValueError("classifiers are not supported")

    elif (ensemble_config["classifier"] ==False) and (ensemble_config["regressor"] == True):
        logger.info("predicting regressor")
        target_no_space = target.replace(' ', '_')
        with open(os.path.join(path_to_param, target_no_space) + '_reg.sav', 'rb') as file:
            m = pickle.load(file)
        with open(os.path.join(path_to_scores, target_no_space) + '_reg.sav', 'rb') as file:
            scoring = pickle.load(file) 
        scores = abs(np.mean(scoring['test_MAE']))


    elif (ensemble_config["classifier"] ==True) and (ensemble_config["regressor"] == True):
        logger.info("predicting zero-inflated regressor")
        target_no_space = target.replace(' ', '_')
        with open(os.path.join(path_to_param, target_no_space) + '_zir.sav', 'rb') as file:
            m = pickle.load(file)
        with open(os.path.join(path_to_scores, target_no_space) + '_zir.sav', 'rb') as file:
            scoring = pickle.load(file)
        scores = abs(np.mean(scoring['test_MAE']))

    elif (ensemble_config["classifier"] ==False) and (ensemble_config["regressor"] == False):
        raise ValueError("Both regressor and classifier are defined as false")

    return(m, scores)


def export_prediction(ensemble_config, m, target, target_no_space, X_predict, X_train, y_train, cv, model_out, n_threads=8):
    """
    Exports model predictions to a NetCDF file.

    Parameters
    ----------
    m : object
        The trained model used for predictions.
    target : str
        The name of the target variable.
    target_no_space : str
        The target variable name with spaces replaced by underscores.
    X_predict : pd.DataFrame of shape (n_points, n_features)
        Features to predict on (e.g., environmental data), where n_points
        is the total 1-d size of the features to predict on 
        (ex. 31881600 for full 180x360x41x12 grid).
    model_out : str
        Path where the predictions should be saved.
    n_threads : int, optional, default=1
        The number of threads to use for parallel prediction.
    """

    if (ensemble_config["classifier"] ==False) and (ensemble_config["regressor"] == True):
        with parallel_backend("loky", n_jobs=n_threads):
            d = pp.estimate_prediction_quantiles(
                m, X_predict=X_predict, X_train=X_train, y_train=y_train, cv=cv
            )["predict_stats"]
            d['mean'] = m.predict(X_predict)
        
        d = d.to_xarray()
        d['target'] = target
        export_path = os.path.join(model_out, target_no_space + ".nc")
        os.makedirs(model_out, exist_ok=True)

        d.to_netcdf(export_path) 
        logger.info(f"finished exporting summary stats to: {export_path}")
        
    elif (ensemble_config["classifier"] ==True) and (ensemble_config["regressor"] == True):
        y_clf = y_train.copy()
        y_clf[y_clf > 0] = 1
        
        optimal_threshold = find_optimal_threshold(m.classifier, X_train, y_clf)

        with parallel_backend("loky", n_jobs=n_threads):
            d_both = pp.estimate_prediction_quantiles(
                m, X_predict=X_predict, X_train=X_train, y_train=y_train, cv=cv, threshold=optimal_threshold
            )
            # Generate classifier and regressor stats
            d_clf = d_both['classifier_predict_stats']
            d_reg = d_both['regressor_predict_stats']


        columns = ["ci95_LL", "ci95_UL"]
        d = pd.DataFrame(d_reg)

        for col in columns:
            d[col] = np.where(d_clf[col] < optimal_threshold, 0, d_reg[col])

        with parallel_backend("loky", n_jobs=n_threads):
            d['mean'] = m.predict(X_predict)

        d_clf = d_clf.to_xarray()
        d_reg = d_reg.to_xarray()
        d = d.to_xarray()
        d_clf['target'] = target
        d_reg['target'] = target
        d['target'] = target

        clf_export_path = os.path.join(model_out, "clf", target_no_space + ".nc")
        reg_export_path = os.path.join(model_out,"reg", target_no_space + ".nc")
        zir_export_path = os.path.join(model_out, target_no_space + ".nc")

        for dir_name in ["", "clf", "reg"]:
            os.makedirs(os.path.join(model_out, dir_name), exist_ok=True)

        d_clf.to_netcdf(clf_export_path) 
        logger.info(f"finished exporting summary stats to: {clf_export_path}")
        d_reg.to_netcdf(reg_export_path) 
        logger.info(f"finished exporting summary stats to: {reg_export_path}")
        d.to_netcdf(zir_export_path) 
        logger.info(f"finished exporting summary stats to: {zir_export_path}")

    else:
        raise ValueError("classifiers are not supported")

    #remove loky tmp data:
    shutil.rmtree(temp_folder, ignore_errors=True)

class ModelPredictor:
    """
    Class to predict outcomes using an ensemble of regression models and export the predictions to a NetCDF file.

    Parameters
    ----------
    X_train : pd.DataFrame of shape (n_samples, n_features)
        Training features used for model fitting.
    y : pd.Series of shape (n_samples,) or (n_samples, n_outputs)
        Target values used for model fitting.
    X_predict : pd.DataFrame of shape (n_points, n_features)
        Features to predict on, where `n_points` represents the total number of prediction points 
        (e.g., 31881600 for a full 180x360x41x12 grid).
    model_config : dict
        Dictionary containing model configuration parameters, including:
            - seed : int
                Random seed for reproducibility.
            - root : str
                Path to the root folder.
            - path_out : str
                Directory where predictions are saved.
            - target : str
                File name of the target list.
            - verbose : int
                Verbosity level (0-3).
            - n_threads : int
                Number of threads to use for parallel processing.
            - cv : int
                Number of cross-validation folds.
            - ensemble_config : dict
                Configuration for the ensemble setup, containing:
                    - classifier : bool
                        Whether to train a classification model.
                    - regressor : bool
                        Whether to train a regression model.
                    - m{n} : str
                        Model names (e.g., "m1: 'rf'", "m2: 'xgb'").
    n_jobs : int, optional, default=1
        Number of threads to use for parallel processing.

    Attributes
    ----------
    path_out : str
        Path where predictions and model outputs are saved.
    target : str
        Name of the target variable.
    target_no_space : str
        Target variable name with spaces replaced by underscores.
    verbose : int
        Verbosity level for logging.
    n_jobs : int
        Number of parallel threads used for prediction and cross-validation.

    Methods
    -------
    make_prediction()
        Train the ensemble models and generate predictions, exporting them to NetCDF.
    """

    def __init__(self, X_train, y, X_predict, model_config, n_jobs=1):
        """
        Initialize the `predict` class with training data, prediction data, and model configurations.

        Parameters
        ----------
        X_train : pd.DataFrame of shape (n_samples, n_features)
            Training features used for model fitting.
        y : pd.Series of shape (n_samples,) or (n_samples, n_outputs)
            Target values used for model fitting.
        X_predict : pd.DataFrame of shape (n_points, n_features)
            Features for which predictions are to be made.
        model_config : dict
            Dictionary containing configuration parameters for the model and ensemble.
        n_jobs : int, optional, default=1
            Number of threads for parallel processing.

        Returns
        -------
        None
        """
                
        self.st = time.time()

        self.y = y.sample(frac=1, random_state=model_config['seed']) #shuffle
        self.X_train = X_train.sample(frac=1, random_state=model_config['seed']) #shuffle

        self.seed = model_config['seed']
        self.target = y.name
        self.target_no_space = self.target.replace(' ', '_')
        self.verbose = model_config['verbose']

        self.path_out = os.path.join(model_config['root'], model_config['path_out'], model_config['run_name'])

            
        if model_config['stratify']==True:
            if model_config['upsample']==True:
                self.cv = UpsampledZeroStratifiedKFold(n_splits=model_config['cv'])
                logger.info("upsampling = True")
            else:
                self.cv = ZeroStratifiedKFold(n_splits=model_config['cv'])
        else:
            self.cv = KFold(n_splits=model_config['cv'])

        self.X_predict = X_predict
        X_predict = None
        self.ensemble_config = model_config['ensemble_config']
        self.model_config = model_config

        self.n_jobs = n_jobs

        if (self.ensemble_config["classifier"] ==True) and (self.ensemble_config["regressor"] == False):
            raise ValueError("classifiers are not supported")
        elif (self.ensemble_config["classifier"] ==False) and (self.ensemble_config["regressor"] == False):
            raise ValueError("classifier and regressor can't both be False")
        else:
            self.scoring = {
                'R2': 'r2',
                'MAE': 'neg_mean_absolute_error',
                'RMSE': 'neg_root_mean_squared_error'
            }


        if (self.ensemble_config["regressor"] !=True) and (self.ensemble_config["regressor"] !=False):
            raise ValueError("regressor should be True or False")


        if self.model_config['ensemble_config']['classifier'] and not self.model_config['ensemble_config']['regressor']:
            raise ValueError("classifiers are not supported")
        elif self.model_config['ensemble_config']['classifier'] and self.model_config['ensemble_config']['regressor']:
            self.extension = "_zir.sav"
        else:
            self.extension = "_reg.sav"

        logger.info("initialized prediction")
        
    def make_prediction(self):
        """
        Fit models in the ensemble and generate predictions.

        Predictions are exported to NetCDF files. If the ensemble contains multiple models, 
        predictions are made for each individual model and the ensemble.

        Returns
        -------
        None

        Notes
        -----
        - Individual model predictions and ensemble predictions are saved separately.
        - Performance metrics (e.g., cross-validation scores) are saved for the ensemble.
        - Only regression models are supported; classification is not implemented.
        """

        number_of_models = len(self.ensemble_config) -2
        logger.info(f"number of models in ensemble: {number_of_models}")

        if number_of_models==1:

            m, mae1 = load_model_and_scores(self.path_out, self.ensemble_config, 0, self.target_no_space)

            model_name = self.ensemble_config["m" + str(1)]
            model_out = os.path.join(self.path_out, "predictions", model_name)
            export_prediction(self.ensemble_config, m, self.target, self.target_no_space, self.X_predict, self.X_train, self.y, self.cv, 
                              model_out, n_threads=self.n_jobs)

        elif number_of_models >=2:
                    
            # iteratively make prediction for each model
            models = []
            mae_values = []
            w = []

            for i in range(number_of_models):
                m, mae = load_model_and_scores(self.path_out, self.ensemble_config, i, self.target_no_space)
                model_name = self.ensemble_config["m" + str(i + 1)]
                model_out = os.path.join(self.path_out, "predictions", model_name)

                if (self.ensemble_config["classifier"] ==False) and (self.ensemble_config["regressor"] == True):
                    export_prediction(self.ensemble_config, m, self.target, self.target_no_space, self.X_predict, self.X_train, self.y, self.cv, 
                                    model_out, n_threads=self.n_jobs)
                if (self.ensemble_config["classifier"] ==True) and (self.ensemble_config["regressor"] == True):

                    export_prediction(self.ensemble_config, m, self.target, self.target_no_space, self.X_predict, self.X_train, self.y, self.cv, 
                                     model_out, n_threads=self.n_jobs)

                models.append((model_name, m))
                mae_values.append(mae)

            w = inverse_weighting(mae_values) 
            if (self.ensemble_config["classifier"] ==False) and (self.ensemble_config["regressor"] == True):
                m = VotingRegressor(estimators=models, weights=w).fit(self.X_train, self.y)   
                model_out = os.path.join(self.path_out, "predictions", "ens")
                export_prediction(self.ensemble_config, m, self.target, self.target_no_space, self.X_predict, self.X_train, self.y, self.cv, 
                                model_out, n_threads=self.n_jobs)
                
                #export model object:
                base_output_path = os.path.join(self.path_out, "model", "ens")
                #make new dir if needed
                os.makedirs(base_output_path, exist_ok=True)


                file_path = os.path.join(base_output_path, f"{self.target_no_space}{self.extension}")

                with open(file_path, 'wb') as f:
                    pickle.dump(m, f)

            elif (self.ensemble_config["classifier"] ==True) and (self.ensemble_config["regressor"] == True):
                clf_models = []
                reg_models = []

                for i in range(number_of_models):
                    path_to_param  = os.path.join(self.path_out, "model", self.ensemble_config["m"+str(i+1)])

                    with open(os.path.join(path_to_param, self.target_no_space) + '_clf.sav', 'rb') as file:
                        m_clf = pickle.load(file)       
                    clf_models.append(('clf'+str(i+1), m_clf))  

                    with open(os.path.join(path_to_param, self.target_no_space) + '_reg.sav', 'rb') as file:
                        m_reg = pickle.load(file)

                    reg_models.append(('reg'+str(i+1), m_reg))  

                y = self.y.values.ravel().copy()

                voting_reg = VotingRegressor(estimators=reg_models, weights=w).fit(self.X_train, y)
                y_clf = y.copy()
                y_clf[y_clf > 0] = 1

                voting_clf = VotingClassifier(estimators=clf_models, weights=w, voting="soft").fit(self.X_train, y_clf)

                optimal_threshold = find_optimal_threshold(voting_clf, self.X_train, y_clf)
                m = ZeroInflatedRegressor(
                    classifier=voting_clf,
                    regressor=voting_reg,
                    threshold=optimal_threshold
                )
                m.fit(self.X_train, y)
                model_out = os.path.join(self.path_out, "predictions", "ens")
                export_prediction(self.ensemble_config, m, self.target, self.target_no_space, self.X_predict, self.X_train, self.y, self.cv, 
                                model_out, n_threads=self.n_jobs)
                base_output_path = os.path.join(self.path_out, "model", "ens")

                #make new dir if needed
                os.makedirs(base_output_path, exist_ok=True)

                file_path = os.path.join(base_output_path, f"{self.target_no_space}{self.extension}")

                with open(file_path, 'wb') as f:
                    pickle.dump(m, f)

            else:
                raise ValueError("classifiers are not supported")

            scores = cross_validate(m, self.X_train, self.y, cv=self.cv, verbose=self.verbose, 
                                    scoring=self.scoring, n_jobs=self.n_jobs)

            model_out_scores = os.path.join(self.path_out, "scoring", "ens")

            #make new dir if needed
            os.makedirs(model_out_scores, exist_ok=True)


            scores_file_path = os.path.join(model_out_scores, f"{self.target_no_space}{self.extension}")

            with open(scores_file_path, 'wb') as f:
                pickle.dump(scores, f)

        else:
            raise ValueError("at least one model should be defined in the ensemble")

        et = time.time()
        elapsed_time = et-self.st
        logger.info("finished")
        logger.info(f"execution time: {elapsed_time} seconds")



# %%
from operator import index
import numpy as np
import xarray as xr
import xesmf as xe

path = "/home/mv23682/Documents/Abil/studies/wiseman2024/env_data_processing/raw_data/RS_PAR/RS_PAR_ESM-based_fill_monthly_clim_1998-2022_0-200.nc"
name = "PAR"

#load netcdf as xarray
ds1 = xr.open_dataset(path)
ds1 = ds1.drop_vars(['DOI'])
ds1 = ds1.rename(name_dict = {'latitude':'lat', 'longitude':'lon', 'Time':'time', 'Depth':'depth'})

# setup new grid to regrid to:
ds_out = xr.Dataset({'lat': (['lat'], np.arange(-90, 90, 1)),
                    'lon': (['lon'], np.arange(-180, 180, 1)),
                    'depth': (['depth'], np.arange(0, 200, 5)),
                    'time': (['time'], np.arange(1, 13, 1)),
                    })

regridder1 = xe.Regridder(ds1, ds_out, 'conservative', periodic=True)
dr1_out = regridder1(ds1['PAR'],skipna=True, na_thres=0.75)

df1 = dr1_out.to_dataframe(name=name).reset_index()
df1.set_index(['time', 'depth', 'lat', 'lon'], inplace=True)
print(df1)
ds = df1.to_xarray()
print('saving ' + name)
ds.to_netcdf("/home/mv23682/Documents/Abil/studies/wiseman2024/env_data_processing/regridded_data/" + name + ".nc")

print("fin")
# %%



import numpy as np
import pandas as pd
import xarray as xr
import logging
logger = logging.getLogger("abil")

from sklearn.datasets import make_regression
from sklearn.utils import resample
from sklearn.metrics import roc_curve, roc_auc_score

from joblib import delayed
import warnings
from xgboost import DMatrix, Booster
from sklearn.exceptions import NotFittedError
from xgboost import XGBClassifier, XGBRegressor, DMatrix
from sklearn.pipeline import Pipeline
from sklearn.compose import TransformedTargetRegressor

def do_nothing(x):
    """
    Apply no transformation to the input values.

    Parameters
    ----------
    x : array-like
        Input values.

    Returns
    -------
    y : array-like
        Non-transformed values.
    """
    return(x)

def is_xgboost_model(model):
    """
    Recursively check if the model is an XGBoost model,
    even if it's wrapped in a Pipeline or TransformedTargetRegressor.
    Uses `getattr` to check for XGBoost-specific attributes.
    """
    # Base case: Check if the model is directly an XGBoost model
    if isinstance(model, (XGBClassifier, XGBRegressor)):
        return True
    # Check for XGBoost-specific attributes using `getattr`
    elif getattr(model, "get_booster", None) is not None or getattr(model, "booster", None) is not None:
        return True

    # Recursive case: Unwrap the model if it's a wrapper
    if isinstance(model, Pipeline):
        # Get the final estimator in the pipeline
        return is_xgboost_model(model.steps[-1][1])
    elif isinstance(model, TransformedTargetRegressor):
        # Get the regressor inside the TransformedTargetRegressor
        return is_xgboost_model(model.regressor)
    elif getattr(model, "estimator", None) is not None:
        # Handle other meta-estimators (e.g., GridSearchCV, BaggingRegressor)
        return is_xgboost_model(model.estimator)
    elif getattr(model, "base_estimator", None) is not None:
        # Handle ensemble models (e.g., AdaBoost, Bagging)
        return is_xgboost_model(model.base_estimator)

    # If none of the above, it's not an XGBoost model
    return False

def xgboost_get_n_estimators(model):
    """
    Recursively extract the `n_estimators` parameter from an XGBoost model,
    even if it's wrapped in a Pipeline or TransformedTargetRegressor.
    """
    # Unwrap TransformedTargetRegressor
    if isinstance(model, TransformedTargetRegressor):
        model = model.regressor

    # Unwrap Pipeline
    if isinstance(model, Pipeline):
        model = model.steps[-1][1]  # Get the final estimator

    # Check if the model is an XGBoost model and has `n_estimators`
    if isinstance(model, (XGBClassifier, XGBRegressor)):
        return model.n_estimators
    elif hasattr(model, "n_estimators"):
        return model.n_estimators

    # If no `n_estimators` is found, raise an error or return a default value
    raise ValueError("Could not extract `n_estimators` from the model.")
from sklearn.pipeline import Pipeline
from sklearn.compose import TransformedTargetRegressor
from xgboost import XGBClassifier, XGBRegressor
from sklearn.exceptions import NotFittedError
from sklearn.base import _is_fitted

def _predict_one_member(i, member, chunk, proba=False, threshold=0.5):
    """
    """
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', category=UserWarning)
        if proba:
            if isinstance(member, Booster):
                # For XGBoost Booster, use predict() to get probabilities
                proba_preds = member.predict(DMatrix(chunk, feature_names=chunk.columns.tolist()), iteration_range=(i, i+1))
                # For binary classification, proba_preds is already the probability of the positive class
                positive_proba = proba_preds

                if (positive_proba < 0).any() or (positive_proba > 1).any():
                    raise ValueError("Probabilities are outside the valid range [0, 1]")
                
            else:
                # For other models, use predict_proba()
                proba_preds = member.predict_proba(chunk)
                
                # Extract probabilities for the positive class (second column)
                positive_proba = proba_preds[:, 1]
            
            # Convert probabilities to binary predictions based on the threshold
            return (positive_proba > threshold).astype(int)
        else:
            if isinstance(member, Booster):
                prediction = member.predict(DMatrix(chunk, feature_names=chunk.columns.tolist()), iteration_range=(0, i+1))
            #    if (prediction > 1).any():
            #        raise ValueError("prediction >1 was made :)")
            else:
                prediction =member.predict(chunk)
            return prediction


def upsample(d, target, ratio=10):
    """
    Upsample zero and non-zero observations in the dataset to balance classes.
    
    Parameters
    ----------
    d : pd.DataFrame
        Input dataframe.
    target : str
        Target column for upsampling.
    ratio : int, default=10
        Ratio of zeros to non-zero samples after upsampling.

    Returns
    -------
    ix : pd.DataFrame
        Upsampled dataframe.
    """
    # Separate presence (non-zero) and absence (zero) indices
    presence_indices = d[d[target] != 0].index
    absence_indices = d[d[target] == 0].index

    # Calculate the number of zero samples to match the ratio
    num_presence = len(presence_indices)
    num_absence_to_sample = num_presence * ratio

    # Sample zeros with replacement if needed
    sampled_absence_indices = np.random.choice(absence_indices, size=num_absence_to_sample, replace=True)

    # Create the upsampled dataframe by combining presences and sampled absences
    upsampled_df = pd.concat([d.loc[presence_indices], d.loc[sampled_absence_indices]], ignore_index=True)

    return upsampled_df


def merge_obs_env(obs_path="../data/gridded_abundances.csv",
                  env_path="../data/env_data.nc",
                  env_vars=None,
                  out_path="../data/obs_env.csv"):
    """
    Merge observational and environmental datasets based on spatial and temporal indices.

    Parameters
    ----------
    obs_path : str, default="../data/gridded_abundances.csv"
        Path to observational data CSV.
    env_path : str, default="../data/env_data.nc"
        Path to environmental data NetCDF file.
    env_vars : list of str, optional
        List of environmental variables to include in the merge.
    out_path : str, default="../data/obs_env.csv"
        Path to save the merged dataset.

    Returns
    -------
    None
    """
    if env_vars is None:
        env_vars = ["temperature", "sio4", "po4", "no3", "o2", "mld", "DIC",
                    "TA", "irradiance", "chlor_a", "Rrs_547", "Rrs_667", "CI_2",
                    "time", "depth", "lat", "lon"]

    d = pd.read_csv(obs_path)
    d = d.convert_dtypes()
    d = d.groupby(['Latitude', 'Longitude', 'Depth', 'Month']).mean().reset_index()
    d.rename({'Latitude': 'lat', 'Longitude': 'lon', 'Depth': 'depth', 'Month': 'time'}, inplace=True, axis=1)
    d.set_index(['lat', 'lon', 'depth', 'time'], inplace=True)

    logger.info("loading env")
    ds = xr.open_dataset(env_path)
    logger.info("converting to dataframe")
    df = ds.to_dataframe()
    ds = None
    df.reset_index(inplace=True)
    df = df[env_vars]
    df.set_index(['lat', 'lon', 'depth', 'time'], inplace=True)
    logger.info("merging environment")
    out = d.merge(df, how="left", left_index=True, right_index=True)
    out.to_csv(out_path, index=True)
    logger.info("fin")

def example_data(
    y_name, 
    n_samples=100, 
    n_features=5, 
    noise=0.1, 
    train_to_predict_ratio=0.7, 
    zero_to_non_zero_ratio=0.5, 
    random_state=59
    ):
    """
    Generate training and prediction datasets with ['lat', 'lon', 'depth', 'time'] indices.
    Includes zeros in the target and allows upsampling of zero values.
    
    Parameters:
        y_name (str): Name of the target variable.
        n_samples (int): Total number of samples to generate (training + prediction).
        n_features (int): Number of features for the dataset.
        noise (float): Noise level for the regression data.
        train_to_predict_ratio (float): Ratio of training to prediction data.
        zero_to_non_zero_ratio (float): Ratio of zero to non-zero target values after upsampling.
        random_state (int): Random seed for reproducibility.

    Returns:
        X_train (pd.DataFrame): Training feature dataset with MultiIndex.
        X_predict (pd.DataFrame): Prediction feature dataset with MultiIndex.
        y (pd.Series): Target variable for training dataset.
    """
    np.random.seed(random_state)  # Set random seed for reproducibility
    
    # Generate regression data
    X, y = make_regression(n_samples=n_samples, n_features=n_features, noise=noise, random_state=random_state)
    
    # Introduce zeros randomly in the target variable
    zero_fraction = 0.2  # Fraction of zero values to introduce
    zero_indices = np.random.choice(len(y), size=int(len(y) * zero_fraction), replace=False)
    y[zero_indices] = 0
    
    # Ensure target values are non-negative by taking absolute values
    y = np.abs(y)
    
    # Split data into training and prediction sets
    train_size = int(n_samples * train_to_predict_ratio)
    X_train, X_predict = X[:train_size], X[train_size:]
    y_train = y[:train_size]
    
    # Upsample zeros in the training set
    X_train_df = pd.DataFrame(X_train, columns=[f"feature_{i+1}" for i in range(X_train.shape[1])])
    y_train_series = pd.Series(y_train, name=y_name)
    
    zero_mask = y_train_series == 0
    non_zero_mask = y_train_series > 0
    
    X_zeros = X_train_df[zero_mask]
    y_zeros = y_train_series[zero_mask]
    X_non_zeros = X_train_df[non_zero_mask]
    y_non_zeros = y_train_series[non_zero_mask]
    
    # Upsample zeros to achieve the desired zero_to_non_zero_ratio
    upsample_count = int(len(y_non_zeros) * zero_to_non_zero_ratio)
    X_zeros_upsampled, y_zeros_upsampled = resample(
        X_zeros, y_zeros, 
        replace=True, 
        n_samples=upsample_count, 
        random_state=random_state
    )
    
    # Combine upsampled data
    X_train_combined = pd.concat([X_non_zeros, X_zeros_upsampled], axis=0)
    y_train_combined = pd.concat([y_non_zeros, y_zeros_upsampled], axis=0)
    
    # Shuffle combined data
    combined_indices = np.random.permutation(X_train_combined.index)
    X_train_combined = X_train_combined.loc[combined_indices]
    y_train_combined = y_train_combined.loc[combined_indices]
    
    # Generate random latitude, longitude, depth, and time
    latitudes_train = np.round(np.random.uniform(-90, 90, size=X_train_combined.shape[0]), -1)
    longitudes_train = np.round(np.random.uniform(-180, 180, size=X_train_combined.shape[0]), -1)
    depths_train = np.round(np.random.uniform(0, 200, size=X_train_combined.shape[0]), -1)
    times_train = np.round(np.random.randint(1, 13, size=X_train_combined.shape[0]), -1)
    
    latitudes_predict = np.round(np.random.uniform(-90, 90, size=X_predict.shape[0]), -1)
    longitudes_predict = np.round(np.random.uniform(-180, 180, size=X_predict.shape[0]), -1)
    depths_predict = np.round(np.random.uniform(0, 200, size=X_predict.shape[0]), -1)
    times_predict = np.round(np.random.randint(1, 13, size=X_predict.shape[0]), -1)
    
    # Set MultiIndex for X_train and X_predict
    X_train_combined.index = pd.MultiIndex.from_arrays(
        [latitudes_train, longitudes_train, depths_train, times_train],
        names=['lat', 'lon', 'depth', 'time']
    )
    X_predict = pd.DataFrame(X_predict, columns=[f"feature_{i+1}" for i in range(X_predict.shape[1])])
    X_predict.index = pd.MultiIndex.from_arrays(
        [latitudes_predict, longitudes_predict, depths_predict, times_predict],
        names=['lat', 'lon', 'depth', 'time']
    )
    
    # Set y_train index
    y_train_combined.index = X_train_combined.index
    
    return X_train_combined, X_predict, y_train_combined


def abbreviate_species(species_name):
    """
    Abbreviate a species name by shortening the first word to its initial.

    Parameters
    ----------
    species_name : str
        Full species name.

    Returns
    -------
    str
        Abbreviated species name.
    """
    words = species_name.split()
    if len(words) == 1:
        return species_name
    abbreviated_name = words[0][0].upper() + '.'
    abbreviated_name += ' ' + ' '.join(words[1:])
    return abbreviated_name


def inverse_weighting(values):
    """
    Compute inverse weighting for a list of values.

    Parameters
    ----------
    values : list of float
        Input values.

    Returns
    -------
    list of float
        Normalized inverse weights.
    """
    inverse_weights = [1 / value for value in values]
    total_inverse_weight = sum(inverse_weights)
    normalized_weights = [weight / total_inverse_weight for weight in inverse_weights]
    return normalized_weights


def find_optimal_threshold(model, X, y_test):
    """
    Finds the optimal probability threshold for binary classification using the ROC curve and Youden's Index.

    Parameters:
    -----------
    model : sklearn classifier
        A fitted binary classification model
    X : array-like of shape (n_samples, n_features)
        Input features for the test or validation set.
    y_test : array-like of shape (n_samples,)
        True binary labels for the test or validation set.

    Returns:
    --------
    optimal_threshold : float
        The optimal probability threshold for classifying a sample as present.
    """
    
    # Get predicted probabilities for the positive class
    y_pred_proba = model.predict_proba(X)[:, 1]

    # Compute ROC curve
    fpr, tpr, thresholds = roc_curve(y_test, y_pred_proba)

    # Calculate optimal threshold using Youden's Index
    youden_index = tpr - fpr
    optimal_idx = np.argmax(youden_index)
    optimal_threshold = thresholds[optimal_idx]

    return optimal_threshold


def weighted_quantile(x, weights, q=.5):
    """
    Computes the weighted quantile(s) of a dataset.

    Parameters:
    -----------
    x : array-like of shape (n_samples,)
        The data for which to compute the quantile(s).
    weights : array-like of shape (n_samples,)
        The weights corresponding to each data point in `x`.
    q : float or array-like of floats, default=0.5
        The quantile(s) to compute. Must be between 0 and 1. If an array is provided, 
        the function will return the weighted quantiles for each value in `q`.

    Returns:
    --------
    result : float or list of floats
        The weighted quantile(s) corresponding to the input `q`. If `q` is a single float, 
        the result is a single value. If `q` is an array-like, the result is a list of quantiles."
    """
    # build a dataframe with two columns, "weight" and "data"
    df = pd.DataFrame.from_dict(dict(data=x, weight=weights))
    # sort values ascending based on the "data"
    df.sort_values("data", inplace=True)
    # the cumulative sum of the weight column tells you how much 
    # of the weight in the dataframe is less than or equal to
    # this row. 
    weight_sums = df.weight.cumsum()
    # the sum of all weight is the last value of that cumulative sum    
    total_weight = weight_sums.iloc[-1]
    # then, the quantile value at each row is equal to the weight
    # at or below that row divided by the total weight. 
    observed_quantiles = (weight_sums/total_weight)
    if isinstance(q, float):
        if not ((0 <= q) & (q <= 1)):
            raise ValueError("quantile must be between zero and one")
        # Give me all rows in the data where the fraction of weight
        # smaller than that row is at least the quantile we're looking for. 
        at_or_above_q = df.data[observed_quantiles >= q]
        # and use the first row that is greater than q% of the weight. 
        result = at_or_above_q.iloc[0]
    else:
        result = []
        for q_ in q:
            if not ((0 <= q_) & (q_ <= 1)):
                raise ValueError("all quantiles must be between zero and one")
            at_or_above_q = df.data[observed_quantiles >= q_]
            result.append(at_or_above_q.iloc[0])
    return result


# %%

import numpy as np
import xarray as xr
import xesmf as xe

# Define the file list
files = [
    '20030101_20230131',
    '20030201_20230228',
    '20030301_20220331',
    '20030401_20210430',
    '20030501_20220531',
    '20030601_20220630',
    '20020701_20230731',
    '20020801_20230831',
    '20020901_20230930',
    '20021001_20211031',
    '20021101_20211130',
    '20021201_20221231'
]

base_path = '/home/mv23682/Documents/Abil/studies/wiseman2024/env_data_processing/raw_data/MODIS_Rrs547/'

# Create a target grid for regridding
ds_out = xr.Dataset({
    'lat': (['lat'], np.arange(-90, 90, 1)),
    'lon': (['lon'], np.arange(-180, 180, 1))
})

# Create a filename for storing the weights
weights_file = "regridded_data/regrid_weights.nc"

print('Initializing regridder')
# Open the first dataset to create a persistent regridder (only calculated once)
initial_ds = xr.open_dataset(f"{base_path}AQUA_MODIS.{files[0]}.L3m.MC.RRS.Rrs_547.9km.nc")
regridder = xe.Regridder(initial_ds, ds_out, method="bilinear", periodic=True, filename=weights_file)

ds_all = []

for i in range(0, 12):
    ds = xr.open_dataset(f"{base_path}AQUA_MODIS.{files[i]}.L3m.MC.RRS.Rrs_547.9km.nc")
    ds_out = xr.Dataset({'lat': (['lat'], np.arange(-90, 90, 1)),
                    'lon': (['lon'], np.arange(-180, 180, 1))
                    })
    print(ds)
    # Regrid chlor_a to the new lat/lon grid (with persistent regridder, reuse weights)
    regridder = xe.Regridder(ds, ds_out, method="bilinear", periodic=True, reuse_weights=True, filename=weights_file)
    dr_out = regridder(ds['Rrs_547'])
    
    # Add time and depth dimensions in one step
    dr_out = dr_out.assign_coords(time=i+1).expand_dims(time=[i+1])
    dr_out_depth = dr_out.expand_dims(depth=[0, 205])
    
    # Interpolate along the depth axis from 0 to 305 meters at 5m intervals
    dr_out_depth = dr_out_depth.interp(depth=np.arange(0, 205, 5))
    
    # Append the dataset to the list
    ds_all.append(dr_out_depth)

# Concatenate all datasets along the time dimension
ds = xr.concat(ds_all, dim="time")

# Additionally, fill NaN regions (high latitudes during winter) with min or 10% of maximum, whichever is lower
# If you want to skip this, skip lines 64-89 and swtich "filled_ds" to "ds" for lines 90-96
# Calculate the max and min along the time dimension
max_values = ds.max(dim='time', skipna=True)
print(max_values)
min_values = ds.min(dim='time', skipna=True)
print(min_values)

# Print the sizes of max_values and min_values
print("Size of max_values:", max_values.sizes)
print("Size of min_values:", min_values.sizes)

# Calculate 10% of the max values
ten_percent_max = 0.1 * max_values

# Create a mask for locations with missing data (more than 0 and less than 12 NaNs)
null_counts = ds.isnull().sum(dim='time')
mask = (null_counts > 0) & (null_counts < 12)

# Create an array to hold the fill values
fill_values = xr.where(mask, np.minimum(ten_percent_max, min_values), np.nan)

# Fill missing values only where they are present in the original dataset
filled_ds = ds.where(ds.notnull(), fill_values)

# Ensure not to fill if all values are missing (i.e., keep original data if all months are NaN)
filled_ds = xr.where(null_counts == 12, ds, filled_ds)

# Assign a name to the DataArray
filled_ds.name = 'Rrs_547'

# Save the filled dataset to NetCDF
filled_ds.to_netcdf("/home/mv23682/Documents/Abil/studies/wiseman2024/env_data_processing/regridded_data/Rrs_547.nc")

print("Finished processing and saved to Rrs_547.nc")
# %%